// HTag
//
// basic classes used by WikiParser

#include "Inc.h"

// HTag
//
// store the parameters and clear mNext

HTag::HTag(HTag* prev, TTag type, bool para) : mType(type), mPara(para)
{
	mNext = 0;
	mFD = 0;
	if (prev)	prev->SetSubtag(this);
}

// ~HTag
//
// ensure the children are deleted
// this was probably done when the close tags were written

HTag::~HTag()
{
	DeleteChildren();
}

// DeleteChildren
//
// delete the children of a tag, so the file is ready
// for our changes

void HTag::DeleteChildren()
{
	delete mNext;
	mNext = 0;
}

// WriteTags
//
// delete our children, then write the tags to file

void HTag::WriteTags(const char* tags, ...)
{
	DeleteChildren();

	static char	buffer[4096];

	va_list	args;
	va_start(args, tags);
	vsprintf(buffer, tags, args);
	va_end(args);

	if (mFD && buffer[0])	fprintf(mFD, "%s", buffer);
}

// SetSubtag
//
// set mNext

void HTag::SetSubtag(HTag* tag)
{
	if (tag != mNext)
	{
		DeleteChildren();
		mNext = tag;
		if (mNext)	mNext->SetFD(mFD);
	}
}

// SetFD
//
// set mFD

void HTag::SetFD(FILE *fd)
{
	mFD = fd;
}

// HTagSimple
//
// write open tags and store close tags

HTagSimple::HTagSimple(HTag* prev, TTag type, bool para, const char* open, const char* close) : HTag(prev, type, para)
{
	WriteTags(open);
	mClose = close;
}

// ~HTagSimple
//
// more subtle than it looks - WriteTags will delete anything contained in our tags
// and *then* write our tags.  The HTag destructor doesn't then do anything.

HTagSimple::~HTagSimple()
{
	WriteTags(mClose);
}

// HList
//
// handles list and list element tags

HList::HList(HTag* prev, TTag type, const char* open, const char* close) : HTagSimple(prev, type, true, open, close)
{
	mOpen = false;
}

HList::~HList()
{
	CloseItem();
}

void HList::OpenItem()
{
	CloseItem();
	WriteTags("<li>");
	mOpen = true;
}

void HList::CloseItem()
{
	if (mOpen)	WriteTags("</li>\n");
	mOpen = false;
}

// HCharFmt
//
// remembers the current format

HCharFmt::HCharFmt(HTag* prev, CharFmt fmt) : HTag(prev, TCharFmt, true)
{
	static const char*	tags[4] = { "", "<i>", "<b>", "<b><i>" };

	mFmt = fmt;

	WriteTags(tags[int(fmt)]);
}

HCharFmt::~HCharFmt()
{
	static const char*	tags[4] = { "", "</i>", "</b>", "</i></b>" };

	WriteTags(tags[int(mFmt)]);
}

void HCharFmt::ChangeFormat(CharFmt fmt)
{
	static const char*	tags[4][4] =
	{
		{ "",			"<i>",			"<b>",		"<b><i>" },		// from normal
		{ "</i>",		"",				"</i><b>",	"</i><b><i>" },	// from italic
		{ "</b>",		"</b><i>",		"",			"<i>" },		// from bold
		{ "</i></b>",	"</i></b><i>",	"</i>",		"" },			// from bold italic
	};

	WriteTags(tags[int(mFmt)][int(fmt)]);
	mFmt = fmt;
}

// HHeading
//
// very simple

HHeading::HHeading(HTag* prev, int level) : HTag(prev, THeading, true), mLevel(level)
{
	WriteTags("<h%d>", mLevel);
}

HHeading::~HHeading()
{
	WriteTags("</h%d>\n", mLevel);
}

// HTable
//
// table object

HTable::HTable(HTag* prev, int border, int padding, const char* caption) : HTag(prev, TTable, false)
{
	WriteTags("<table cellpadding=\"%d\" border=\"%d\"><caption>%s</caption>\n", padding, border, caption);
	WriteTags("<tr>\n");

	mCol = 1;
	mRow = 1;
	mCCell = 0;
}

HTable::~HTable()
{
	CloseCell();
	WriteTags("</tr>\n");
	WriteTags("</table>\n");
}

void HTable::OpenCell(int col, int row, bool heading)
{
	if ((col == mCol) && (row == mRow) && mCCell)	return;

	CloseCell();

	while (row > mRow)
	{
		WriteTags("</tr>\n");
		WriteTags("<tr>\n");
		mCol = 1;
		mRow++;
	}
	while (col > mCol)
	{
		WriteTags("<td>\n");
		WriteTags("</td>\n");
		mCol++;
	}

	if (heading)
	{
		WriteTags("<th>\n");
		mCCell = 2;
	}
	else
	{
		WriteTags("<td>\n");
		mCCell = 1;
	}
}

void HTable::CloseCell()
{
	if (mCCell)
	{
		if (mCCell == 1)		WriteTags("</td>\n");
		else if (mCCell == 2)	WriteTags("</th>\n");
		mCol++;
	}
	mCCell = 0;
}
