// HTag
//
// basic classes used by WikiParser

// outline:
//
// as we parse, we travel down the chain
// if the HTag type we want is present:
//   check its parameters match ours
// else
//   delete and replace with our type
// then accept the HTag = move down the chain
// pass to next layer of parser
//
// e.g. > * Hello there
//      --|-|------------> creates a blockquote tag
//        --|------------> creates an unordered list tag
//          -------------> creates a paragraph tag and a text format tag
//
// new HTag prints the open tag text
// other HTag members can print intermediate text
// delete HTag closes the tags

// TTag
//
// tag types

enum TTag
{
	TNone = 0,
	TParagraph,
	TBlockquote,
	TUnorderedList,
	TOrderedList,
	TVirtualList,
	TCenter,
	TPre,
	TTable,
	TCharFmt,
	THeading,
};

// CharFmt
//
// text formats

enum CharFmt
{
	Plain = 0,
	Italic,
	Bold,
	BoldItalic
};

// HTag
//
// represents one or more HTML tags grouped
// together as a logical tag, and containing subtags
// specified in a chain via mNext
//
// overridable functions:
//
// HTag()		this prints the "open" tags for the group
// ~HTag()		this closes all tags for the group
// other		other functions can open/close embedded tags
//				it is up to the user to not SetSubtag() until
//              the tags in this object are ready
//
// WriteTags	should be called in the subclass to write tags to file
//				it works like printf

class HTag
{
public:
	HTag(HTag* prev, TTag type, bool para);
	virtual ~HTag();

	TTag		GetType() const			{ return mType; }
	HTag*		GetNext()				{ return mNext; }
	const HTag*	GetNext() const			{ return mNext; }
	TTag		GetNextType() const		{ return mNext ? mNext->mType : TNone; }
	bool		CanContainText() const	{ return mPara; }

	void		SetSubtag(HTag* tag);
	void		SetFD(FILE* fd);

protected:
	void		WriteTags(const char* tags, ...);	// call to write tags to file

private:
	FILE*		mFD;		// file for output
	HTag*		mNext;		// next HTag in parse tree
	const TTag	mType;		// type field
	const bool	mPara;		// tag can contain text directly without a <p>

	void		DeleteChildren();
};

// HTagSimple
//
// a simple HTag that has fixed strings for open and close

class HTagSimple : public HTag
{
public:
	HTagSimple(HTag* prev, TTag type, bool para, const char* open, const char* close);
	~HTagSimple();

private:
	const char*	mClose;
};

// HParagraph
// HBlockquote
// HVirtualList
// HCenter
// HPre
//
// simple subclasses of HTagSimple

class HParagraph : public HTagSimple
{
public:
	HParagraph(HTag* prev) : HTagSimple(prev, TParagraph, true, "<p>\n", "</p>\n") { }
};

class HBlockquote : public HTagSimple
{
public:
	HBlockquote(HTag* prev) : HTagSimple(prev, TBlockquote, false, "<blockquote>\n", "</blockquote>\n") { }
};

class HVirtualList : public HTagSimple
{
public:
	HVirtualList(HTag* prev) : HTagSimple(prev, TVirtualList, true, "<p>\n", "</p>\n") { }
};

class HCenter : public HTagSimple
{
public:
	HCenter(HTag* prev) : HTagSimple(prev, TCenter, true, "<center>", "</center>\n") { }
};

class HPre : public HTagSimple
{
public:
	HPre(HTag* prev) : HTagSimple(prev, TPre, true, "<pre>", "</pre>\n") { }
};

// HList
//  HUnorderedList
//  HOrderedList

class HList : public HTagSimple
{
public:
	HList(HTag* prev, TTag type, const char* open, const char* close);
	~HList();

	void	OpenItem();
	void	CloseItem();

private:
	bool	mOpen;
};

class HUnorderedList : public HList
{
public:
	HUnorderedList(HTag* prev) : HList(prev, TUnorderedList, "<ul>\n", "</ul>\n") { }
};

class HOrderedList : public HList
{
public:
	HOrderedList(HTag* prev) : HList(prev, TOrderedList, "<ol>\n", "</ol>\n") { }
};

// HCharFmt
//
// remembers how it is set and can be changed
// reasonably efficiently (not optimal)

class HCharFmt : public HTag
{
public:
	HCharFmt(HTag* prev, CharFmt fmt);
	~HCharFmt();

	void	ChangeFormat(CharFmt fmt);

private:
	CharFmt	mFmt;
};

// HHeading

class HHeading : public HTag
{
public:
	HHeading(HTag* prev, int level);
	~HHeading();

private:
	const int	mLevel;
};

// HTable

class HTable : public HTag
{
public:
	HTable(HTag* prev, int border, int padding, const char* caption);
	~HTable();

	int		GetCol() const	{ return mCol; }
	int		GetRow() const	{ return mRow; }

	void	OpenCell(int col, int row, bool heading);
	void	CloseCell();

private:
	int		mCol;		// current column
	int		mRow;		// current row
	int		mCCell;		// 0 for no current cell, 1 for a data cell, 2 for a heading cell
};
