// Page.cpp
//
// page class

#include "Inc.h"

// new Page
//
// copy title, set as "not in memory" and clear the
// list of outbound links

Page::Page(const char* title, int flags)
{
	mTitle = StringDup(title);
	if (flags & PF_NONWIKI)	mSpacedTitle = StringDup(title);
	else					mSpacedTitle = MakeSpacedTitle(title);

	mASCII = NULL;
	mNext = NULL;
	mFlags = flags;
	mSeqNo = 0;
	mCB = 0;
	memset(&mTime, 0, sizeof(mTime));

	mLink = NULL;
	mLinkCnt = 0;
}

// delete Page
//
// release resources - a page cannot be deleted while
// in a list or while its data is loaded

Page::~Page()
{
	assert(!mASCII);
	assert(!mNext);
	delete[] mLink;
	delete[] mTitle;
	delete[] mSpacedTitle;
}

// ModifyFlags
//
// change a page's flags

int Page::ModifyFlags(int on, int off)
{
	on &= ~PF_CHANGED;
	off &= ~PF_CHANGED;

	mFlags &= ~off;
	mFlags |= on;

	return mFlags;
}

// GetDateString
//
// get page modification date as a (static) string

const char* Page::GetDateString() const
{
	return OS_GetDateString(mTime);
}

// GetTimeString
//
// get page modification time as a (static) string

const char* Page::GetTimeString() const
{
	return OS_GetTimeString(mTime);
}

// GetFilename
//
// get filename as a (static) string

const char* Page::GetFilename() const
{
	static char	fn[_MAX_PATH];

	if (mFlags & PF_NONWIKI)		strcpy(fn, mTitle);
	else							sprintf(fn, "%s.wik", mTitle);

	return fn;
}

// ScanForLinks
//
// find all the links in a document - this is a list
// of strings, terminated with NULL

void Page::ScanForLinks()
{
	if (!mASCII)				return;
	if (mFlags & PF_NONWIKI)	return;

	delete[] mLink;
	mLinkCnt = 0;

	char**	array = new char*[32];
	int		ix = 0;
	int		max = 32;

	const char*	doc = mASCII;

	// simple parser uses CS_isalpha, GetAlphaNumeric
	// and IsValidTitle to extract titles from the document
	// the title is indexed regardless of whether or not
	// the page exists
	while (*doc)
	{
		if (CS_isalpha(*doc))
		{
			char*	word = GetAlphaNumeric(doc);
			bool	link = false;

			if (IsValidTitle(word))
			{
				link = true;
			}
			else if (doc[0] == ':')
			{
				if (!stricmp(word, "link"))
				{
					doc++;
					delete[] word;
					word = GetAlphaNumeric(doc);

					if (word[0])	link = true;
				}
				else if (!stricmp(word, "nolink") || !stricmp(word, "nl"))
				{
					doc++;
					delete[] word;
					word = GetAlphaNumeric(doc);
				}
				else if (!stricmp(word, "img"))
				{
					doc++;
					delete[] word;
					word = GetURL(doc);
				}
				else if (cfg.InterWiki.Lookup(word))
				{
					doc++;
					delete[] word;
					word = GetAlphaNumeric(doc);
				}
				else
				{
					doc -= strlen(word);
					delete[] word;
					word = GetURL(doc);
				}
			}

			if (link)
			{
				// is it already in the array?
				int	ii;

				for (ii = 0; ii < ix; ii++)
				{
					if (!stricmp(array[ii], word))	break;
				}
				if (ii == ix)
				{
					// no!
					if (ix == max)
					{
						// extend array
						max *= 2;
						char**	narray = new char*[max];
						for (int ii = 0; ii < ix; ii++)
						{
							narray[ii] = array[ii];
						}
						delete[] array;
						array = narray;
					}
					array[ix++] = word;
					word = 0;
				}
			}

			delete[] word;
		}
		else if (doc[0] == '[' && doc[1] != '\0' && doc[1] == '[') // Support for [[pagename]] links
		{
			// Search for ]]
			int len=0;
			bool	link = false;

			doc += 2;

			while (doc[len] != '\0' && doc[len] != 0xD && doc[len] != ']')
			{
				len++;
			}

			char *temptitle = new char[len+1];
			memcpy(temptitle, doc, len);
			temptitle[len] = '\0';

			if ( doc[len] == ']' && doc[len+1] != '\0' && doc[len+1] == ']' )
			{
				// We have a possible link to a Wikipage
				doc += strlen(temptitle) + strlen("]]");

				Capitalise(temptitle);
				StripSpaces(temptitle);
				
				if (IsValidTitle(temptitle))
				{
					link = true;
				}
				else
				{
					// Invalid WikiTitle, see if it is an acceptable title (alphanumeric)
					if (temptitle[0] && IsAcceptableTitle(temptitle))
					{
						link = true;
					}
				}
			}

			if (link)
			{
				// is it already in the array?
				int	ii;

				for (ii = 0; ii < ix; ii++)
				{
					if (!stricmp(array[ii], temptitle))	break;
				}
				if (ii == ix)
				{
					// no!
					if (ix == max)
					{
						// extend array
						max *= 2;
						char**	narray = new char*[max];
						for (int ii = 0; ii < ix; ii++)
						{
							narray[ii] = array[ii];
						}
						delete[] array;
						array = narray;
					}
					array[ix++] = temptitle;
					temptitle = 0;
				}
			}

			delete[] temptitle;
		}
		else 
		{
			doc++;
		}
	}

	// reallocate
	mLink = new char*[ix];
	mLinkCnt = ix;

	for (int ii = 0; ii < ix; ii++)
	{
		mLink[ii] = array[ii];
	}

	delete[] array;
}
