// PageHTML.cpp
//
// HTML output, other than normal Wiki pages

#include "Inc.h"
#ifndef SANOS
#include <direct.h>
#endif

// WriteFromTemplate
//
// write from template to file until we hit a $$BODY$$, then return true
// return false when finished

static void Write(const char* sp, int len, FILE* fd)
{
	for (int ii = 0; ii < len; ii++)
	{
		if (sp[ii] != 0x0D)		fputc(sp[ii], fd);
	}
}

bool PageSet::WriteFromTemplate(const char*& lp, FILE* fd, const char* title, const char* heading, bool& footer, const UserInfo* user, char** incpagetitle, bool convert) const
{
	for (;;)
	{
		// find first $$
		const char*	sp = lp;

		while (*lp && ((lp[0] != '$') || (lp[1] != '$')))	lp++;

		Write(sp, lp - sp, fd);

		if (!*lp)	return false;

		// find second $$
		sp = lp;
		lp += 2;
		
		while (*lp && ((lp[0] != '$') || (lp[1] != '$')))	lp++;

		if (!*lp)
		{
			Write(sp, lp - sp, fd);
			return false;
		}

		sp += 2;

		char*	value = new char[lp - sp + 1];
		memcpy(value, sp, lp - sp);
		value[lp - sp] = 0;

		lp += 2;

		if (!value[0])
		{
			// $$$$ escapes $$
			fprintf(fd, "$$");
		}
		else if (!stricmp(value, "BODY"))
		{
			if (!incpagetitle)
			{
				delete[] incpagetitle;
				incpagetitle= NULL;
			}

			delete[] value;
			footer = false;
			return true;
		}
		else if (!stricmp(value, "FOOTER"))
		{
			delete[] value;
			footer = true;
			return true;
		}
		else if (!strnicmp(value, "PAGE", 4))
		{
			// Insert a Wiki page

			// Figure out the pagetitle
			int pagetitlelength = strlen(&value[5]); 
			char*	pagetitle = new char[pagetitlelength + 1]; // Reserve one extra byte for '\0'

			strncpy( pagetitle, &value[5], pagetitlelength);
			pagetitle[pagetitlelength] = '\0'; // Make sure the pagetitle is terminated

			// See if it is a good title
			if (IsAcceptableTitle(pagetitle))
			{
				// Let "them" know we have a pagetitle
				if (!incpagetitle)
				{
					delete[] incpagetitle;
					incpagetitle= NULL;
				}

				*incpagetitle = StringDup(pagetitle);
			}
			else
			{
				// No good title
				char	fn[_MAX_PATH];
				sprintf(fn, "%s\\%s", cfg.Pathname, pagetitle);

				// Do we want to upload ?????
				FILE*	fd2 = fopen(fn, "rb");

				if (!fd2)
				{
					fprintf(fd, "%s<a href=\"/UploadFile?file=%s\">&gt;</a>", pagetitle, pagetitle);
				}
				else
				{
					fclose(fd2);
					fprintf(fd, "<pre>%s</pre>", pagetitle);
				}
			}

			delete[] pagetitle;
			delete[] value;
			return true;
		}
		else if (!strnicmp(value, "REVERSELINK", 11))
		{
			// Create a bulleted list of all pages that link to the page provided in the params
			char*	pagetitle;

			if (value[11] == ':')
			{
				// Figure out the pagetitle if one is provided in the $$ tag
				int pagetitlelength = strlen(&value[12]) ;
				pagetitle = new char[pagetitlelength + 1]; // One extra byte for '\0'
				strncpy( pagetitle, &value[12], pagetitlelength);
				pagetitle[pagetitlelength] = '\0';
			}
			else
			{
				// Figure out what page is currently written so we know what to reverselink
				pagetitle = new char[strlen(title) + 1]; // One extra byte for '\0'
				strcpy(pagetitle, title);
				StripSpaces(pagetitle);
			}
		
			Page* tempPage;

			tempPage = cfg.Pages->FindPage(pagetitle);

			if (tempPage)
			{
				cfg.Pages->ReverseLinkList(tempPage, fd, user, convert);
			}

			delete[] pagetitle;

		}
		else if (!stricmp(value, "TITLE"))
		{
			fprintf(fd, "%s", title);
		}
		else if (!stricmp(value, "HOMEPAGE"))
		{
			fprintf(fd, user->GetHomepage());
		}
		else if (!stricmp(value, "DEFPAGE"))
		{
			fprintf(fd, cfg.DefaultPage);
		}
		else if (!stricmp(value, "HEADING"))
		{
			fprintf(fd, "%s", heading);
		}
		else if (!stricmp(value, "VERSION"))
		{
			fprintf(fd, "%s", WIKISERVER_VERSION);
		}
		else
		{
			fprintf(fd, "$$%s$$", value);
		}

		delete[] value;
	}
}


// WriteEditPage
//
// write the HTML for an edit page to the file provided

int PageSet::WriteEditPage(const char* pname, int nrows, int ncols, FILE* fd, const UserInfo* user) const
{
	Page*	page = FindPage(pname);

	if (!page && !IsAcceptableTitle(pname))
	{
		page = FindSysPage(pname);

		if (!page)	return 400;
	}

	if (page && (page->GetFlags() & PF_META))							return 400;
	if (page && (page->GetFlags() & PF_ADMINEDIT) && !user->Auth('a'))	return 401;

	char*	spname = page ? StringDup(page->GetSpacedTitle()) : MakeSpacedTitle(pname);

	nrows = nrows ? nrows : 18;
	ncols = ncols ? ncols : 80;

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char	title[1024];
	sprintf(title, "Edit %s", spname);

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, fd, title, spname, footer, user, &incpagetitle, false))  // Editpages are not written to HTML so convert is always false
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(page, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<form method=\"POST\" action=\"SubmitPage\">\n");
				fprintf(fd, "<blockquote>\n");
				fprintf(fd, "<p>\n");
				fprintf(fd, "<input type=\"submit\" value=\" Save \">\n");
				fprintf(fd, "<input type=\"reset\" value=\" Reset \">\n");
				fprintf(fd, "<input type=\"hidden\" name=\"page\" value=\"%s\">\n", pname);
				fprintf(fd, "<input type=\"hidden\" name=\"seqno\" value=\"%d\">\n", page ? page->GetSeqNo() : 0);
				fprintf(fd, "Show <a href=\"WikiFormatting\" target=\"_blank\">Wiki Formatting</a> (opens in a new window)\n");
				fprintf(fd, "</p>\n");
				fprintf(fd, "<p>\n");
				fprintf(fd, "<textarea name=\"text\" rows=%d cols=%d>", nrows, ncols);
				if (page)
				{
					WriteAsTextArea(page, fd);
				}
				else
				{
					fprintf(fd, "Define %s here.\n", pname);
				}
				fprintf(fd, "</textarea>\n");
				fprintf(fd, "</p>\n");
				fprintf(fd, "</blockquote>\n");
				fprintf(fd, "</form>\n");
			}
		}
		else
		{
			fprintf(fd, "<p>\n");
			fprintf(fd, "<a href=\"EditPage?page=%s", pname);
			if (nrows > 20)
			{
				fprintf(fd, "&rows=18&cols=80\">EditPage</a> using a smaller text area.<br>\n");
			}
			else
			{
				fprintf(fd, "&rows=34&cols=96\">EditPage</a> using a larger text area.<br>\n");
			}
			fprintf(fd, "</p>\n");
		}
	}

	delete[] spname;

	return 200;
}

// WriteSearchForm
//
// write the HTML for a search form to the file provided

bool WriteSearchForm(const char* name, const char* value, FILE *fd)
{
	fprintf(fd, "<form method=\"post\" action=\"/\">\n");
	fprintf(fd, "<p>\n");
	fprintf(fd, "<input type=\"text\" size=\"40\" name=\"%s\" value=\"%s\">\n", name, value);
	fprintf(fd, "<input type=\"submit\" value=\" Go \">\n");
	fprintf(fd, "</p>\n");
	fprintf(fd, "</form>\n");

	return true;
}

// WriteGoogleSearchForm
//
// write the HTML for a Google search form to the file provided

bool WriteGoogleSearchForm(const char* value, FILE* fd)
{
	fprintf(fd, "<form method=\"get\" action=\"http://www.google.com/search\">\n");
	fprintf(fd, "<p>\n");
	fprintf(fd, "<input type=\"text\" size=\"40\" name=\"q\" maxlength=\"255\" value=\"%s\">\n", value);
	fprintf(fd, "<input type=\"submit\" value=\" Go \">\n");
	fprintf(fd, "</p>\n");
	fprintf(fd, "</form>\n");

	return true;
}

// WriteAmazonSearchForm
//
// write the HTML for an Amazon search form to the file provided

bool WriteAmazonSearchForm(const char* value, FILE* fd)
{
	fprintf(fd, "<form method=\"post\" action=\"http://www.amazon.com/exec/obidos/search-handle-form/102-5867411-7491324\">\n");
	fprintf(fd, "<p>\n");
	fprintf(fd, "<input type=\"hidden\" name=\"index\" value=\"All Products\">\n");
	fprintf(fd, "<input type=\"text\" size=\"40\" name=\"field-keywords\" value=\"%s\">\n", value);
	fprintf(fd, "<input type=\"submit\" name=\"Go\" value=\" Go \"\n");
	fprintf(fd, "</p>\n");
	fprintf(fd, "</form>\n");

	return true;
}

// WriteMergePage
//
// write the HTML for a merge page to the file provided

bool PageSet::WriteMergePage(Page* page, const char* newv, FILE* fd, const UserInfo* user) const
{
	const char*	pname = page->GetTitle();
	const char* spname = page->GetSpacedTitle();

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char	title[1024];
	sprintf(title, "Merge %s", spname);

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, fd, title, spname, footer, user, &incpagetitle, false)) // Mergepage is not written to HTML so convert is always false
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(page, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>Another user edited the page while you were working.</p>\n");
				fprintf(fd, "<p>The current page is shown in the left window; the page you submitted is shown in the right window.</p>\n");
				fprintf(fd, "<p>Please merge your changes from the right window into the left window, and resave.</p>\n");

				fprintf(fd, "<form method=\"POST\" action=\"SubmitPage\">\n");
				fprintf(fd, "<blockquote>\n");
				fprintf(fd, "<p>\n");
				fprintf(fd, "<input type=\"submit\" value=\" Save \">\n");
				fprintf(fd, "<input type=\"reset\" value=\" Reset \">\n");
				fprintf(fd, "<input type=\"hidden\" name=\"page\" value=\"%s\">\n", pname);
				fprintf(fd, "<input type=\"hidden\" name=\"seqno\" value=\"%d\">\n", page->GetSeqNo());
				fprintf(fd, "</p>\n");
				fprintf(fd, "<p>\n");
				fprintf(fd, "<textarea name=\"text\" rows=%d cols=%d>", 18, 40);
				WriteAsTextArea(page, fd);
				fprintf(fd, "</textarea>\n");
				fprintf(fd, "<textarea name=\"newv\" rows=%d cols=%d>", 18, 40);
				WriteAsTextArea(newv, fd);
				fprintf(fd, "</textarea>\n");
				fprintf(fd, "</p>\n");
				fprintf(fd, "</blockquote>\n");
				fprintf(fd, "</form>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	return true;
}

// WriteEditedMessage
//
// write the HTML for the thank-you message to the file provided

bool PageSet::WriteEditedMessage(Page* page, FILE* outfd, const UserInfo* user) const
{
	const char*	title = page->GetTitle();
	const char*	spname = page->GetSpacedTitle();

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char	ptitle[1024];
	sprintf(ptitle, "Thank you for editing %s", spname);

	char	heading[1024];

	if (page->GetFlags() & PF_NONWIKI)
	{
		strcpy(heading, spname);
	}
	else
	{
		sprintf(heading, "<a href=\"ReverseLink?page=%s\">%s</a>", title, spname);
	}

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, outfd, ptitle, heading, footer, user, &incpagetitle, false)) // Thank you page is not written to HTML so convert is always false
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(page, incpagetitle, outfd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				// Write thank you
				if (page->GetFlags() & PF_NONWIKI)
				{
					fprintf(outfd, "<p><h1>Thank you for editing %s.</h1>Return to <a href=\"AdminLogin\">AdminLogin</a>.</h1></p>\n", spname);
				}
				else
				{
					fprintf(outfd, "<p><h1>Thank you for editing <a href=\"%s\">%s</a>.</h1></p>\n", title, spname);
				}
			}
		}
		else
		{
			WriteFooter(0, outfd, user);
		}
	}

	return true;
}

// WriteDeletedMessage
//
// write the HTML for the thank-you message to the file provided

bool PageSet::WriteDeletedMessage(const char* title, FILE* outfd, const UserInfo* user) const
{
	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char	ptitle[1024];
	sprintf(ptitle, "Page %s deleted", title);

	char		heading[1024];
	sprintf(heading, "Page %s deleted", title);

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, outfd, ptitle, heading, footer, user, &incpagetitle, false)) // Deleted message is never written to HTML so convert is always false
	{
		if (!footer)
		{	
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, outfd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(outfd, "<p><h1>Page %s<a href=\"%s\">?</a> deleted.</h1></p>\n", title, title);
			}
		}
		else
		{
			WriteFooter(0, outfd, user);
		}
	}

	return true;
}

// WriteAsTextArea
//
// write text to an HTML file encoded for correct reproduction

bool PageSet::WriteAsTextArea(const char* text, FILE* outfd) const
{
	if (text)
	{
		while (char ch = *text++)
		{
			if (ch == '&')			fprintf(outfd, "&amp;");
			else if (ch == '\"')	fprintf(outfd, "&quot;");
			else if	(ch == '<')		fprintf(outfd, "&lt;");
			else if (ch == '>')		fprintf(outfd, "&gt;");
			else if (ch == 0x0D)	;
			else					fputc(ch, outfd);
		}
	}

	return true;
}

// WriteAsTextArea
//
// write a page as a text area to the HTML file provided

bool PageSet::WriteAsTextArea(Page* page, FILE* outfd) const
{
	LoadPage(page);

	bool rc = WriteAsTextArea(page->mASCII, outfd);

	UnloadPage(page);

	return rc;
}

// ListAllPages
//
// make the HTML for the ListAllPages page

void PageSet::ListAllPages(FILE* fd, const UserInfo* user, const bool convert) const
{
	PageList*	list = MakeAlphabeticList();

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;
	char*		incpagetitle = NULL;

	while (WriteFromTemplate(lp, fd, "List All Pages", "List All Pages", footer, user, &incpagetitle, convert))
	{
		if (!footer)
		{	
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user, convert);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following is a complete list of pages on this server:</p>\n");
				fprintf(fd, "<ul>\n");

				for (int ii = 0; ii < list->cnt; ii++)
				{
					if (list->list[ii]->GetFlags() & PF_META)	continue;

					if (!convert)
					{
						fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list->list[ii]->GetTitle(), list->list[ii]->GetTitle());
					}
					else
					{
						fprintf(fd, "<li><a href=\"%s.html\">%s</a></li>\n", list->list[ii]->GetTitle(), list->list[ii]->GetTitle());
					}
				}
				fprintf(fd, "</ul>");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete list;
}

// RecentChanges
//
// make the HTML for the RecentChanges page

void PageSet::RecentChanges(FILE* fd, const UserInfo* user, const bool convert) const
{
	PageList*	list = MakeRecentList();

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, fd, "Recent Changes", "Recent Changes", footer, user, &incpagetitle, false))
	{
		if (!footer)
		{	
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user, convert);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following pages have recently been edited:</p>\n");
				bool	intab = false;

				time_t	ctime = 0;

				int	ndays = 7;
				
				for (int ii = 0; ii < list->cnt; ii++)
				{
					time_t ftime = list->list[ii]->GetTime();

					if (!ftime)	continue;

					if (!OS_IsSameDay(ctime, list->list[ii]->GetTime()))
					{
						ndays--;
						if (!ndays)	break;

						if (intab)	fprintf(fd, "</table></blockquote>\n");

						fprintf(fd, "<p><b>%s</b></p>\n", list->list[ii]->GetDateString());
						fprintf(fd, "<blockquote><table>\n");
						
						intab = true;

						ctime = list->list[ii]->GetTime();
					}

					if (!convert)
					{
						fprintf(fd, "<tr><td>%s : </td><td><a href=\"%s\">%s</a></td><td></td></tr>\n",
							list->list[ii]->GetTimeString(), 
							list->list[ii]->GetTitle(), 
							list->list[ii]->GetTitle());
					}
					else
					{
						fprintf(fd, "<tr><td>%s : </td><td><a href=\"%s.html\">%s</a></td><td></td></tr>\n",
							list->list[ii]->GetTimeString(), 
							list->list[ii]->GetTitle(), 
							list->list[ii]->GetTitle());
					}
				}
				if (intab)	fprintf(fd, "</table></blockquote>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete list;
}

// ReverseLink
//
// Make the HTML for a ReverseLink page

void PageSet::ReverseLink(Page* page, FILE* fd, const UserInfo* user) const
{
	// make list of pages
	PageList*	list = MakeAlphabeticList();

	// for each page, look for this title in its list
	for (int ii = 0; ii < list->cnt; ii++)
	{
		Page*	pp = list->list[ii];
		int		jj;

		for (jj = 0; jj < pp->mLinkCnt; jj++)
		{
			if (!stricmp(pp->mLink[jj], page->mTitle))	break;
		}
		if (jj == pp->mLinkCnt)
		{
			// no link - clear list element
			list->list[ii] = NULL;
		}
	}

	// print HTML
	const char*	title = page->GetTitle();
	char*		stitle = MakeSpacedTitle(title);

	char	htitle[1024];
	sprintf(htitle, "Reverse Link - %s", stitle);

	char	heading[1024];
	sprintf(heading, "Reverse Link - <a href=\"%s\">%s</a>", title, stitle);

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	// TODO get value of convert from someplace
	while (WriteFromTemplate(lp, fd, htitle, heading, footer, user, &incpagetitle, false))
	{
		if (!footer)
		{	
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(page, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following pages link to <a href=\"%s\">%s</a>:</p>\n", title, title);
				fprintf(fd, "<ul>\n");

				bool	any = false;

				for (int ii = 0; ii < list->cnt; ii++)
				{
					if (!list->list[ii])						continue;
					if (list->list[ii]->GetFlags() & PF_META)	continue;

					fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list->list[ii]->GetTitle(), list->list[ii]->GetTitle());

					any = true;
				}

				fprintf(fd, "</ul>\n");
				if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete[] stitle;
	delete list;
}

// ReverseLinkList
//
// Make only a list of links

void PageSet::ReverseLinkList(const Page* page, FILE* fd, const UserInfo* user, const bool convert) const
{
	// make list of pages
	PageList*	list = MakeAlphabeticList();

	// for each page, look for this title in its list
	for (int ii = 0; ii < list->cnt; ii++)
	{
		Page*	pp = list->list[ii];
		int		jj;

		for (jj = 0; jj < pp->mLinkCnt; jj++)
		{
			if (!stricmp(pp->mLink[jj], page->mTitle))	break;
		}

		if (jj == pp->mLinkCnt || pp->mTitle == page->mTitle)
		{
			// no link - clear list element
			list->list[ii] = NULL;
		}
	}

	// print HTML
	const char*	title = page->GetTitle();
	char*		stitle = MakeSpacedTitle(title);

	bool	any = false;

	for (int iii = 0; iii < list->cnt; iii++)
	{
		if (!list->list[iii])						continue;
		if (list->list[iii]->GetFlags() & PF_META)	continue;

		if (convert)
		{
			fprintf(fd, "<a href=\"%s.html\">%s</a><br>\n", list->list[iii]->GetTitle(), list->list[iii]->GetTitle());
		}
		else
		{
			fprintf(fd, "<a href=\"%s\">%s</a><br>\n", list->list[iii]->GetTitle(), list->list[iii]->GetTitle());
		}

		any = true;
	}
			
	delete[] stitle;
	delete list;
}

// TitleSearch
//
// Make the HTML for a TitleSearch results page

void PageSet::TitleSearch(const char* search, FILE* fd, const UserInfo* user) const
{
	// make list of pages
	PageList*	list = MakeAlphabeticList();

	char	htitle[1024];
	sprintf(htitle, "Title Search - %s", search);

	char	heading[1024];
	sprintf(heading, "<a href=\"TitleSearch\">Title Search</a> - %s", search);

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	// TODO get value of convert from someplace
	while (WriteFromTemplate(lp, fd, htitle, heading, footer, user, &incpagetitle, false))
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following pages were found:</p>\n");
				fprintf(fd, "<ul>\n");

				// for each page, look for the search string
				bool	any = false;

				for (int ii = 0; ii < list->cnt; ii++)
				{
					Page*	pp = list->list[ii];
					
					if (pp->GetFlags() & PF_META)	continue;

					if (MatchWildcardString(pp->GetTitle(), search))
					{
						fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", pp->GetTitle(), pp->GetTitle());
						any = true;
					}
				}

				fprintf(fd, "</ul>\n");
				if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete list;
}

// FullSearch
//
// Make the HTML for a FullSearch results page

void PageSet::FullSearch(const char* search, FILE* fd, const UserInfo* user) const
{
	// make list of pages
	PageList*	list = MakeAlphabeticList();

	char	htitle[1024];
	sprintf(htitle, "Full Search - %s", search);

	char	heading[1024];
	sprintf(heading, "<a href=\"FullSearch\">Full Search</a> - %s", search);

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	// TODO get value of convert from someplace
	while (WriteFromTemplate(lp, fd, htitle, heading, footer, user, &incpagetitle, false))
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following pages were found:</p>\n");
				fprintf(fd, "<ul>\n");

				// for each page, load it in and look for the search string
				bool	any = false;

				for (int ii = 0; ii < list->cnt; ii++)
				{
					Page*	pp = list->list[ii];

					if (pp->GetFlags() & PF_META)	continue;

					bool match = MatchWildcardString(pp->GetTitle(), search);

					if (!match)
					{
						LoadPage(pp);
						match = MatchWildcardString(pp->mASCII, search);
						UnloadPage(pp);
					}

					if (match)
					{
						fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", pp->GetTitle(), pp->GetTitle());
						any = true;
					}
				}		

				fprintf(fd, "</ul>\n");
				if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete list;
}

// ListOrphans
//
// Make the HTML for the ListOrphans page

void PageSet::ListOrphans(FILE* fd, const UserInfo* user, const bool convert) const
{
	// make list of pages
	PageList*	list = MakeAlphabeticList();

	// for each page, set its PF_ORPHANED flag
	int	ii;

	for (ii = 0; ii < list->cnt; ii++)	list->list[ii]->mFlags |= PF_ORPHANED;

	// for each page, clear the orphaned flag of all pages in its link list
	for (ii = 0; ii < list->cnt; ii++)
	{
		Page*	pp = list->list[ii];

		for (int jj = 0; jj < pp->mLinkCnt; jj++)
		{
			Page*	pp2 = FindPage(pp->mLink[jj]);

			if (pp2 == pp)	continue;	// a link to itself doesn't count

			if (pp2)	pp2->mFlags &= ~PF_ORPHANED;
		}
	}

	// now only pages which are orphaned have their flag set

	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	while (WriteFromTemplate(lp, fd, "List Orphans", "List Orphans", footer, user, &incpagetitle, convert))
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user, convert);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>The following pages are orphaned - that is, no pages in this Wiki link to them:</p>\n");
				fprintf(fd, "<ul>\n");

				bool	any = false;

				for (ii = 0; ii < list->cnt; ii++)
				{
					if (!(list->list[ii]->GetFlags() & PF_ORPHANED))	continue;
					list->list[ii]->mFlags &= ~PF_ORPHANED;

					if (list->list[ii]->GetFlags() & PF_META)			continue;
					if (list->list[ii] == mWikiFooter)					continue;	// can't be orphaned!
					if (list->list[ii] == mWikiSystemFooter)			continue;	// can't be orphaned!

					fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list->list[ii]->GetTitle(), list->list[ii]->GetTitle());

					any = true;
				}

				fprintf(fd, "</ul>\n");
				if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}

	delete list;
}

// AdminLogin
//
// Make the HTML for the AdminLogin page

void PageSet::AdminLogin(FILE* fd, const UserInfo* user) const
{
	const char*	lp = mPageTemplate->mASCII;
	bool		footer = false;

	char*       incpagetitle = NULL;

	// TODO get value of convert from someplace
	while (WriteFromTemplate(lp, fd, "Admin Login", "Admin Login", footer, user, &incpagetitle, false))
	{
		if (!footer)
		{
			if (incpagetitle)
			{
				// Write the included Wikipage
				WriteIncludedPage(NULL, incpagetitle, fd, user);
				delete[] incpagetitle;
				incpagetitle= NULL;
			}
			else
			{
				fprintf(fd, "<p>Please select a command from the list:</p>\n");
				fprintf(fd, "<ul>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=Config.wkc\">Edit basic configuration</a> (Config.wkc)</li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=Users.wkc\">Edit user list</a> (Users.wkc)</li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=MimeTypes.wkc\">Edit MIME type list</a> (MIMETypes.wkc)</li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=InterWiki.wkc\">Edit InterWiki list</a> (InterWiki.wkc)</li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=LockPages.wkc\">Edit locked pages list</a> (LockPages.wkc)</li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=WikiFooter\">Edit WikiFooter</a></li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=WikiSystemFooter\">Edit WikiSystemFooter</a></li>\n");
				fprintf(fd, "<li><a href=\"EditPage?page=PageTemplate.html\">Edit HTML template</a> (PageTemplate.html)</li>\n");
				fprintf(fd, "<li><a href=\"DumpAllPages\">Dump All Pages</a> (backup the Wiki)</li>\n");
				fprintf(fd, "<li><a href=\"AuthorLogin\">Author Login</a> (admin logout)</li>\n");
				fprintf(fd, "</ul>\n");
				fprintf(fd, "<ul>\n");
				fprintf(fd, "<li><a href=\"AdminLogin?cmd=convert\">Convert all Wiki files to HTML</a></li>\n");
				fprintf(fd, "<li><a href=\"AdminLogin?cmd=shutdown\">Shutdown the server</a></li>\n");
				fprintf(fd, "</ul>\n");
			}
		}
		else
		{
			WriteFooter(0, fd, user);
		}
	}
}


void ScanDirCB2(const char* dirname, const char* filename, const char* basename, const char* ext, void* apikey);

// ConvertPages
//
// Do page conversion

void PageSet::ConvertPages(FILE* fd, const UserInfo* user) const
{
	PageList*	list = MakeAlphabeticList();

	// Create HTML folder (With dat and time to make folder unique)
	time_t ltime;
	char HTMLDir[_MAX_PATH];

	time(&ltime);
	sprintf(HTMLDir, "HTML - %s", OS_GetDateString(ltime), OS_GetTimeString(ltime));
	_mkdir(HTMLDir);

	// Copy all files to HTML folder except for wik & wkc files
	OS_ScanDir(cfg.Pathname, ::ScanDirCB2, &HTMLDir);

	// Process WikiPages	
	for (int ii = 0; ii < list->cnt; ii++)
	{
		char	fn[MAX_PATH];

		sprintf(fn, "%s\\%s.html", HTMLDir, list->list[ii]->mTitle);

		if (list->list[ii]->mFlags & PF_META)
		{
			// Check if it is a Metapage that we want to convert
			if ( !( !strcmp (list->list[ii]->mTitle, "ListAllPages")
				  || !strcmp(list->list[ii]->mTitle, "ListOrphans")
				  || !strcmp(list->list[ii]->mTitle, "RecentChanges")) )
			{
				continue;
			}
				  
		    FILE*	fd = fopen(fn, "w");
			if (!fd) continue;

			// Call the appropriate function for each metapage that is allowed to be converted
			if (!strcmp(list->list[ii]->mTitle, "ListAllPages"))  ListAllPages(fd, user, true);
			if (!strcmp(list->list[ii]->mTitle, "ListOrphans"))   ListOrphans(fd, user, true);
			if (!strcmp(list->list[ii]->mTitle, "RecentChanges")) RecentChanges(fd, user, true);

			fclose(fd);
		}
		else
		{
			FILE*	fd = fopen(fn, "w");
			if (!fd) continue;

			WriteAsHTML(list->list[ii], fd, user, true);

			fclose(fd);
		}

	}

	delete list;

	AdminLogin(fd, user);
}


void ScanDirCB2(const char* dirname, const char* filename, const char* basename, const char* ext, void* apikey)
{
	if (!basename[0])	return;

	char	fn[_MAX_PATH];

	// Do the following for all files except .wik and .wkc files.
	if (!(!strcmp(ext, "wik") || !strcmp(ext, "wkc")))
	{
		sprintf(fn, "copy \"%s\\%s\" \"%s\\%s\" >nul", dirname, filename, apikey, filename);
		system(fn);
	}
}
