// PageSet.cpp
//
// a set of pages

#include "Inc.h"

#define	RADIX_SIZE	(27*27)

static void ScanDirCB(const char* dirname, const char* filename, const char* basename, const char* ext, void* apikey);

// new PageSet
//
// create a page set from the files in a directory

PageSet::PageSet()
{
	mPageRadix = new Page*[RADIX_SIZE];
	memset(mPageRadix, 0, sizeof(Page*)*RADIX_SIZE);

	mPageCnt = 0;
	mWikiFooter = 0;
	mWikiSystemFooter = 0;
	mPageTemplate = 0;
	mWkcConfig = 0;
	mWkcInterWiki = 0;
	mWkcLockPages = 0;
	mWkcMimeTypes = 0;
	mWkcUsers = 0;

	// load database
	OS_ScanDir(cfg.Pathname, ::ScanDirCB, this);

	// now merge in dump file
	FILE*	fd = fopen("DumpAllPages", "rb");
	if (fd)
	{
		AddFromDump(fd);
		fclose(fd);
		remove("DumpAllPages.bak");
		rename("DumpAllPages", "DumpAllPages.bak");
	}

	// create default WikiFooter
	EnsurePage("WikiFooter",		STD_FOOTER);
	EnsurePage("WikiSystemFooter",	STD_SYSFOOTER);

	// create default template file
	EnsureSysPage("PageTemplate.html", STD_HTML);

	// create default config files
	EnsureSysPage("Config.wkc",		STD_CONFIG);
	EnsureSysPage("InterWiki.wkc",	STD_INTERWIKI);
	EnsureSysPage("LockPages.wkc",	STD_LOCKPAGES);
	EnsureSysPage("MimeTypes.wkc",	STD_MIMETYPES);
	EnsureSysPage("Users.wkc",		STD_USERS);
}

// delete PageSet
//
// unload all pages and delete them

PageSet::~PageSet()
{
	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		Page*	pp = mPageRadix[ii];

		while (pp)
		{
			Page*	next = pp->mNext;
			UnloadPage(pp, true);
			pp->mNext = NULL;
			delete pp;
			pp = next;
		}
	}
}

// EnsurePage
// EnsureSysPage
//
// defines a page if it doesn't exist

void PageSet::EnsurePage(const char* name, const char* def)
{
	Page*	pp = FindPage(name);

	if (!pp)
	{
		pp = CreatePage(name);
		pp->mASCII = StringDup(def);

		pp->mFlags |= PF_CHANGED;
		UnloadPage(pp);
	}

	if (pp->mCB)	(this->*(pp->mCB))(pp);
}

void PageSet::EnsureSysPage(const char* name, const char* def)
{
	Page*	pp = FindSysPage(name);

	if (!pp)
	{
		pp = CreateSysPage(name);
		pp->mASCII = StringDup(def);

		pp->mFlags |= PF_CHANGED;
		UnloadPage(pp);
	}

	if (pp->mCB)	(this->*(pp->mCB))(pp);
}

// ScanDirCB
//
// called back from OS_ScanDir to load database

static void ScanDirCB(const char* dirname, const char* filename, const char* basename, const char* ext, void* apikey)
{
	PageSet*	ps = (PageSet*)apikey;
	ps->ScanDirCB(dirname, filename, basename, ext);
}

void PageSet::ScanDirCB(const char* dirname, const char* filename, const char* basename, const char* ext)
{
	if (!basename[0])	return;

	char	fn[_MAX_PATH];

	sprintf(fn, "%s\\%s", dirname, filename);

	if (!stricmp(ext, "wik"))
	{
		Page*	pp = CreatePage(basename);

		pp->mTime = OS_GetModifiedTime(fn);

		LoadPage(pp);
		pp->ScanForLinks();
		UnloadPage(pp);
	}
	else if (!stricmp(ext, "wkc") || !stricmp(filename, "PageTemplate.html"))
	{
		Page*	pp = CreateSysPage(filename);

		pp->mTime = OS_GetModifiedTime(fn);

		LoadPage(pp);
	}
}

// encode
//
// encode a character into it's alphabetic index, 0 to 25,
// else 26
// this isn't ISO compliant but that's ok, it just means
// names that start with ISO alpha characters are hashed
// together

static int encode(char c)
{
	if ((c >= 'a') && (c <= 'z'))	return c - 'a';
	if ((c >= 'A') && (c <= 'Z'))	return c - 'A';
	return 26;
}

// MakeRadix
//
// make a radix for the page from the first two letters
// of its title

int	PageSet::MakeRadix(const char* title)
{
	assert(RADIX_SIZE == 27*27);

	if (title[0])
	{
		return encode(title[0]) + 27 * encode(title[1]);
	}

	return RADIX_SIZE - 1;
}

// FindPage
//
// find a named page in the set

Page* PageSet::FindPage(const char* title) const
{
	if (!IsAcceptableTitle(title))	return NULL;

	return FindSysPage(title);
}

// FindSysPage
//
// find a page which might be a system page

Page* PageSet::FindSysPage(const char* title) const
{
	int radix = MakeRadix(title);

	Page*	pp;

	for (pp = mPageRadix[radix]; pp; pp = pp->mNext)
	{
		if (!stricmp(title, pp->mTitle))	return pp;
	}

	return NULL;
}

// CreatePage
//
// find a named page in the set; if it does not exist, create it

Page* PageSet::CreatePage(const char* title, int flags)
{
	if (!IsAcceptableTitle(title))	return NULL;

	Page*	pp = FindPage(title);

	if (pp)	return pp;

	return _CreatePage(title, flags);
}

// CreateSysPage
//
// find a system page in the set; if it does not exist, create it

Page* PageSet::CreateSysPage(const char* title, int flags)
{
	Page*	pp = FindSysPage(title);

	if (pp)	return pp;

	return _CreatePage(title, flags | PF_KEEP | PF_ADMINEDIT | PF_NONWIKI);
}

// _CreatePage
//
// create a page in the set

Page* PageSet::_CreatePage(const char* title, int flags)
{
	int	radix = MakeRadix(title);

	Page* pp = new Page(title, flags);
	pp->mNext = mPageRadix[radix];
	mPageRadix[radix] = pp;

	// see if this is a locked page
	if (!strcmp(title, "WikiFooter"))
	{
		mWikiFooter = pp;
		pp->mFlags |= PF_KEEP;
	}
	else if (!strcmp(title, "WikiSystemFooter"))
	{
		mWikiSystemFooter = pp;
		pp->mFlags |= PF_KEEP;
	}
	else if (!strcmp(title, "PageTemplate.html"))
	{
		mPageTemplate = pp;
		pp->mFlags |= PF_KEEP;
	}
	else if (!strcmp(title, "Config.wkc"))
	{
		mWkcConfig = pp;
		pp->mCB = PageSet::UpdateConfig;
	}
	else if (!strcmp(title, "InterWiki.wkc"))
	{
		mWkcInterWiki = pp;
		pp->mCB = PageSet::UpdateInterWiki;
	}
	else if (!strcmp(title, "LockPages.wkc"))
	{
		mWkcLockPages = pp;
		pp->mCB = PageSet::UpdateLockPages;
	}
	else if (!strcmp(title, "MimeTypes.wkc"))
	{
		mWkcMimeTypes = pp;
		pp->mCB = PageSet::UpdateMimeTypes;
	}
	else if (!strcmp(title, "Users.wkc"))
	{
		mWkcUsers = pp;
		pp->mCB = PageSet::UpdateUsers;
	}

	const char*	lock = cfg.LockPages.Lookup(title, 0);

	if (lock)	pp->mFlags |= PF_ADMINEDIT;

	mPageCnt++;

	return pp;
}

// Update*
//
// callbacks for when the config changes

void PageSet::UpdateConfig(Page* pp)
{
	cfg.LoadConfig(mWkcConfig->GetData());
}

void PageSet::UpdateInterWiki(Page* pp)
{
	cfg.LoadInterWiki(mWkcInterWiki->GetData());
}

void PageSet::UpdateLockPages(Page* pp)
{
	cfg.LoadLockPages(mWkcInterWiki->GetData());

	// set ADMINEDIT on wiki pages according to list
	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		for (Page* pp = mPageRadix[ii]; pp; pp = pp->mNext)
		{
			if (!(pp->mFlags & PF_NONWIKI))
			{
				const char*	lock = cfg.LockPages.Lookup(pp->mTitle, 0);

				if (lock)	pp->mFlags |= PF_ADMINEDIT;
				else		pp->mFlags &= ~PF_ADMINEDIT;
			}
		}
	}
}

void PageSet::UpdateMimeTypes(Page* pp)
{
	cfg.LoadMimeTypes(mWkcInterWiki->GetData());
}

void PageSet::UpdateUsers(Page* pp)
{
	cfg.LoadUsers(mWkcUsers->GetData());
}

// GetPathname
//
// get the pathname for a page as a (static) string

const char* PageSet::GetPathname(Page* pp) const
{
	static char	fn[_MAX_PATH];

	sprintf(fn, "%s\\%s", cfg.Pathname, pp->GetFilename());

	return fn;
}

// DeletePage
//
// delete a page

bool PageSet::DeletePage(Page* pp)
{
	if (pp->mFlags & PF_KEEP)	return false;
	if (pp->mFlags & PF_META)	return false;

	const char*	fn = GetPathname(pp);

	// unlink page
	int	radix = MakeRadix(pp->mTitle);

	Page*	page = mPageRadix[radix];

	if (page == pp)
	{
		mPageRadix[radix] = pp->mNext;
	}
	else
	{
		while (page)
		{
			if (page->mNext == pp)
			{
				page->mNext = pp->mNext;
				break;
			}
			page = page->mNext;
		}
	}

	pp->mNext = 0;
	UnloadPage(pp, true);
	delete pp;

	remove(fn);

	return true;
}

// LoadPage
//
// load a page into RAM

bool PageSet::LoadPage(Page* page) const
{
	if (page->mASCII)			return true;

	page->mASCII = LoadFile(GetPathname(page));

	if (!page->mASCII)	return false;

	return true;
}

// UnloadPage
//
// unload a page - optionally force unloading of
// pages which fail to unload or which stay in RAM

bool PageSet::UnloadPage(Page* page, bool force) const
{
	bool	rc = true;

	if (!page->mASCII)			return rc;

	if (page->mFlags & PF_CHANGED)
	{
		if (!WritePage(page) && !force)	return false;
	}

	if ((page->mFlags & PF_KEEP) && !force)	return false;

	delete[] page->mASCII;
	page->mASCII = NULL;

	return rc;
}

// WritePage
//
// write a page to disk, and refresh its links list

bool PageSet::WritePage(Page* page) const
{
	assert(page->mASCII);

	const char*	fn = GetPathname(page);

	FILE*	fd = fopen(fn, "wb");
	if (!fd)	return false;

	page->mFlags &= ~PF_CHANGED;
	fwrite(page->mASCII, 1, strlen(page->mASCII), fd);
	fclose(fd);

	page->mTime = OS_GetModifiedTime(fn);

	page->ScanForLinks();

	return true;
}

// EditPage
//
// set the contents of a page and flush it to disk

void PageSet::EditPage(Page* page, const char* src, size_t srccnt) const
{
	delete[] page->mASCII;
	page->mASCII = new char[srccnt + 1];
	memcpy(page->mASCII, src, srccnt);
	page->mASCII[srccnt] = 0;

	page->mSeqNo++;
	page->mFlags |= PF_CHANGED;

/* Illegal cast in VS 2005, doesn't seem to do anything.

	if (page->mCB)
	{
		(this->*(page->mCB))(page);
	}
*/
	UnloadPage(page);
}

// MakeSortedList
//
// make a list of pages, sorted using the supplied compare function
// this list must eventually be delete[]d

PageList*	PageSet::MakeSortedList(int (*compare)(const Page** pp1, const Page** pp2)) const
{
	PageList*	array = new PageList(mPageCnt);
	Page**		aptr = array->list;

	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		for (Page* pp = mPageRadix[ii]; pp; pp = pp->mNext)
		{
			if (!(pp->mFlags & PF_NONWIKI))		*aptr++ = pp;
		}
	}
	
	array->cnt = aptr - array->list;

	qsort(array->list, array->cnt, sizeof(Page*), (int (*)(const void*, const void*))compare);

	return array;
}

// MakeAlphabeticList
//
// make a list sorted using CS_Compare on the page title

static int AlphaCompare(const Page** pp1, const Page** pp2)
{
	return CS_Compare((*pp1)->GetTitle(), (*pp2)->GetTitle(), true);
}

PageList*	PageSet::MakeAlphabeticList() const
{
	return MakeSortedList(AlphaCompare);
}

// MakeRecentList
//
// make a list sorted from most recently-edited backwards in time

static int TimeCompare(const Page** pp1, const Page** pp2)
{
	time_t	t1;
	time_t	t2;

	t1 = (*pp1)->GetTime();
	t2 = (*pp2)->GetTime();

	if (t1 < t2)	return +1;
	if (t1 > t2)	return -1;

	return 0;
}

PageList*	PageSet::MakeRecentList() const
{
	return MakeSortedList(TimeCompare);
}

// DumpAllPages
//
// dump all the pages to a binary file

static const char hdr_old[13] = "WikiFileDump";
static const char hdr_old2[13] = "WikiFileDmp2";
static const char hdr[13] = "WikiFileDmp3";

int PageSet::DumpAllPages(FILE* fd, const UserInfo* user) const
{
	if (!user->Auth('a'))	return 401;

	fwrite(hdr, 1, 12, fd);

	size_t	len;

	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		for (Page* pp = mPageRadix[ii]; pp; pp = pp->mNext)
		{
			if (pp->GetFlags() & PF_META)	continue;

			LoadPage(pp);

			len = strlen(pp->GetTitle());
			fwrite(&len, 1, sizeof(size_t), fd);
			fwrite(pp->GetTitle(), 1, len, fd);

			len = strlen(pp->mASCII);
			fwrite(&len, 1, sizeof(size_t), fd);
			fwrite(pp->mASCII, 1, len, fd);

			fwrite(&pp->mTime, 1, sizeof(pp->mTime), fd);
			fwrite(&pp->mFlags, 1, sizeof(pp->mFlags), fd);

			UnloadPage(pp);
		}
	}

	len = 0;
	fwrite(&len, 1, sizeof(size_t), fd);

	return 200;
}

// AddFromDump
//
// load the pages from a binary file

void PageSet::AddFromDump(FILE* fd)
{
	char fhdr[12];
	fread(fhdr, 1, 12, fd);

	bool	no_filetime = false;
	bool	no_flags = false;

	if (!memcmp(hdr, fhdr, 12))
	{
	}
	else if (!memcmp(hdr_old, fhdr, 12))
	{
		no_filetime = true;
		no_flags = true;
	}
	else if (!memcmp(hdr_old2, fhdr, 12))
	{
		no_flags = true;
	}
	else
	{
		return;	// invalid header
	}

	for (;;)
	{
		size_t	len;

		// read title
		fread(&len, 1, sizeof(size_t), fd);
		if (!len)	break;

		char*	name = new char[len + 1];
		fread(name, 1, len, fd);
		name[len] = 0;

		// create page
		Page*	pp = CreatePage(name);
		delete[] name;

		// read data
		fread(&len, 1, sizeof(size_t), fd);
		char*	data = new char[len + 1];
		fread(data, 1, len, fd);
		data[len] = 0;

		EditPage(pp, data, len);

		delete[] data;

		if (!no_filetime)
		{
			// read filetime
			time_t	tt;

			fread(&tt, 1, sizeof(tt), fd);

			pp->mTime = tt;
		}

		if (!no_flags)
		{
			// read flags
			fread(&pp->mFlags, 1, sizeof(pp->mFlags), fd);
		}

		if (!no_filetime)
		{
			// set filetime
			OS_SetModifiedTime(GetPathname(pp), pp->mTime);
		}
	}
}
