// PageSet.h
//
// a set of pages

const char gIconName[] = "Icon.jpg";

class PageSet
{
public:
	PageSet();
	~PageSet();

	// basic manipulation
	Page*	FindPage(const char* title) const;
	Page*	CreatePage(const char* title, int flags = 0);
	void	EnsurePage(const char* title, const char* def);
	bool	DeletePage(Page* page);

	// system pages
	Page*	FindSysPage(const char* title) const;
	Page*	CreateSysPage(const char* title, int flags = PF_ADMINEDIT | PF_NONWIKI);
	void	EnsureSysPage(const char* title, const char* def);

	// change contents
	void	EditPage(Page* page, const char* src, size_t srccnt) const;

	// write to/from disk
	bool	LoadPage(Page* page) const;
	bool	UnloadPage(Page* page, bool force = false) const;
	bool	WritePage(Page* page) const;

	// write a Wiki page as HTML
	bool	WriteAsHTML(Page* page, FILE* fd, const UserInfo* user, bool convert = false) const;

	// output HTML for a Wiki string
	void	OutputHTML(const Page* page, const char* text, FILE* fd, const UserInfo* user, bool convert = false) const;

	// output the appropriate footer for page
	void	WriteFooter(const Page* page, FILE* fd, const UserInfo* user) const;

	// output an included Wiki page
	void    WriteIncludedPage(const Page* page, const char* incpagetitle, FILE* outfd, const UserInfo* user, bool convert = false) const;

	// write a correct page link
	void	OutputPageLink(const Page* page, const char* name, FILE* fd, bool convert) const;

	// write from an HTML template into a file; replacing parameters and returning
	// when the client needs to react
	bool	WriteFromTemplate(const char*& lp, FILE* fd, const char* title, const char* heading, bool& footer, const UserInfo* user, char** incpagetitle, const bool convert) const;

	// write an edit page
	int		WriteEditPage(const char* pname, int nrows, int ncols, FILE* fd, const UserInfo* user) const;

	// write a merge page
	bool	WriteMergePage(Page* page, const char* newv, FILE* fd, const UserInfo* user) const;

	// write acknowledgements
	bool	WriteEditedMessage(Page* page, FILE* fd, const UserInfo* user) const;
	bool	WriteDeletedMessage(const char* title, FILE* outfd, const UserInfo* user) const;

	// write as HTML-safe raw text
	bool	WriteAsTextArea(const char* text, FILE* fd) const;
	bool	WriteAsTextArea(Page* page, FILE* fd) const;

	// whole database operations
	int		DumpAllPages(FILE* fd, const UserInfo* user) const;
	void	AddFromDump(FILE* fd);

	// utility pages
	void	ListAllPages(FILE* fd, const UserInfo* user, const bool convert = false) const;
	void	ListOrphans(FILE* fd, const UserInfo* user, const bool convert = false) const;
	void	RecentChanges(FILE* fd, const UserInfo* user, const bool convert = false) const;
	void	ReverseLink(Page* page, FILE* fd, const UserInfo* user) const;

	// Not really a utilitypage, just a hack to get ONLY a list of pages that link to page
	// Usefull for CategoryCategory page listings
	void	ReverseLinkList(const Page* page, FILE* fd, const UserInfo* user, const bool convert) const;

	void	TitleSearch(const char* search, FILE* fd, const UserInfo* user) const;
	void	FullSearch(const char* search, FILE* fd, const UserInfo* user) const;

	void	AdminLogin(FILE* fd, const UserInfo* user) const;

	void	ConvertPages(FILE* fd, const UserInfo* user) const;

	// make a sorted list
	PageList*	MakeSortedList(int (*compare)(const Page** pp1, const Page** pp2)) const;
	PageList*	MakeAlphabeticList() const;
	PageList*	MakeRecentList() const;

	// ONLY for use by ScanDirCB in Page.cpp
	void	ScanDirCB(const char* dirname, const char* filename, const char* basename, const char* ext);

	// callbacks for special pages
	void	UpdateConfig(Page* pp);
	void	UpdateInterWiki(Page* pp);
	void	UpdateLockPages(Page* pp);
	void	UpdateMimeTypes(Page* pp);
	void	UpdateUsers(Page* pp);


private:
	Page**		mPageRadix;
	int			mPageCnt;
	Page*		mWikiFooter;
	Page*		mWikiSystemFooter;
	Page*		mPageTemplate;
	Page*		mWkcConfig;
	Page*		mWkcInterWiki;
	Page*		mWkcLockPages;
	Page*		mWkcMimeTypes;
	Page*		mWkcUsers;

	static int MakeRadix(const char* title);

	Page*		_CreatePage(const char* title, int flags);

	const char*	GetPathname(Page* pp) const;

};

class PageList
{
public:
	PageList(int maxsize)	{ list = new Page*[maxsize]; cnt = 0; }
	~PageList()				{ delete[] list; }

	Page**	list;
	int		cnt;
};

// ouput a search box
bool	WriteSearchForm(const char* name, const char* value, FILE* fd);
bool	WriteGoogleSearchForm(const char* value, FILE* fd);
bool	WriteAmazonSearchForm(const char* value, FILE* fd);
