// WikiParser.cpp
//
// Text -> HTML generation

#include "Inc.h"

// WikiToHTML
//
// WikiParser's only public member

void WikiParser::WikiToHTML(FILE* fd, const char* wiki, CharFmt fmt, const Page* pageref, const UserInfo* user, bool convert)
{
	// initialize the object
	mRoot = new HTag(0, TNone, false);
	mFD = fd;
	mPage = pageref;
	mUser = user;
	mFmt = fmt;
	mAllowAnchor = true;
	mAllowTags = true;
	mConvert = convert;

	mRoot->SetFD(fd);

	while (wiki && wiki[0])
	{
		// get next line
		char*		line = GetLine(wiki);
		const char*	lp = line;

		mAllowAnchor = true;	// allow an anchor definition
		bool	done = MatchTree(lp);

		if (!done)
		{
			// check there's a text-accepting item
			if (!mCT->CanContainText())
			{
				// we need to add a paragraph to this
				if (mCT->GetNextType() != TParagraph)	new HParagraph(mCT);
				mCT = mCT->GetNext();
			}

			// write the HTML
			WriteLine(lp);
		}

		delete[] line;
	}

	delete mRoot;
}

// CharOut
//
// output a safe char

void WikiParser::CharOut(char ch)
{
	switch (ch)
	{
	case '&':	fprintf(mFD, "&amp;");	break;
	case '\"':	fprintf(mFD, "&quot;");	break;
	case '<':	fprintf(mFD, "&lt;");	break;
	case '>':	fprintf(mFD, "&gt;");	break;
	case '\n':	fprintf(mFD, "<br>");	break;
	case 0x0D:							break;
	default:	fputc(ch, mFD);			break;
	}
}

// OutputPageLink
//
// output a page link

void WikiParser::OutputPageLink(const char* pname)
{
	cfg.Pages->OutputPageLink(mPage, pname, mFD, mConvert);
}

// MatchTree
//
// match the parse tree to the commands in the current line

bool WikiParser::MatchTree(const char*& lp)
{
	mCT = mRoot;

	// match outer block items:
	int		minspaces = 0;

	for (;;)
	{
		// check for a blank line, etc.
		if (AcceptLineCommand(lp))	return true;

		// accept left-align command
		const char*	sp = lp;

		if (minspaces)
		{
			// we want to skip a single space (if present) when
			// considering left-aligned commands
			if (lp[0] == ' ')	lp++;
		}

		if (AcceptLeftAlignCommand(lp))
		{
			// after a command, IF we have a space we call AcceptSpaceCommand
			// with a "minimum spaces for pre" of 1 - this means that
			// > text is not preformatted
			// > * Hello is a list
			// >  text is preformatted
			minspaces = 1;
		}
		else
		{
			lp = sp;
		}

		// check for a blank line, etc.
		if (AcceptLineCommand(lp))	return true;

		// accept ` command
		if (AcceptHeading(lp))		return false;

		// accept space command
		if (lp[0] == ' ')
		{
			lp++;
			if (!AcceptSpaceCommand(lp, minspaces))
			{
				lp--;
			}
			else
			{
				mAllowAnchor = false;	// no anchor definition in a list or table

				if (mCT->GetType() == TTable)
				{
					// after a table, we go around again
					// again we set minspace to 1 so if there is NO left align command, but
					// there IS a table/list/etc. command, it is accepted
					minspaces = 1;
					continue;
				}
			}
		}

		// got block-level format
		return false;
	}
}

// AcceptLineCommand
//
// accept a blank line or ----; commands which finish the input line

bool WikiParser::AcceptLineCommand(const char* lp)
{
	if (AcceptPlain(lp))
	{
		return true;
	}

	while (lp[0] == ' ')	lp++;

	if ((lp[0] == '-') && (lp[1] == '-') && (lp[2] == '-') && (lp[3] == '-'))
	{
		mCT->SetSubtag(0);
		fprintf(mFD, "<hr>\n");
		return true;
	}

	//HACK: accept {{{ and }}} and put the <pre> tag
	if ((lp[0] == '{') && (lp[1] == '{') && (lp[2] == '{'))
	{
		mCT->SetSubtag(0);
		fprintf(mFD, "<pre>\n");
		mAllowTags = false;
		return true;
	}
	if ((lp[0] == '}') && (lp[1] == '}') && (lp[2] == '}'))
	{
		mCT->SetSubtag(0);
		fprintf(mFD, "</pre>\n");
		mAllowTags = true;
		return true;
	}

	if (!lp[0])
	{
		// close off the current paragraph, if it's a paragraph
		// otherwise, delete all the stuff we've defined and return
		// to base level (relative to mCT)
		mCT->SetSubtag(0);
		return true;
	}

	if ((lp[0] == '[') && (lp[1] == '?'))
	{
		lp += 2;

		char*	word = GetAlphaNumeric(lp);

		if ((*lp != ']') || !AcceptFormCommand(word))
		{
			lp -= strlen(word) + 2;
			delete[] word;
			return false;
		}

		lp++;
		delete[] word;
		return true;
	}

	if ((lp[0] == '[') && (lp[1] == '!'))
	{
		lp += 2;

		char*	word = GetURL(lp);

		if (*lp != ']')
		{
			lp -= strlen(word) + 2;
			delete[] word;
			return false;
		}

		mCT->SetSubtag(0);
		AcceptFileInsertion(word);

		lp++;
		delete[] word;
		return true;
	}

	return false;
}

// AcceptHeading
//
// accepting a heading

bool WikiParser::AcceptHeading(const char*& lp)
{
	const char* sp = lp;

	while (sp[0] == ' ')	sp++;

	if ((sp[0] == '`') && (sp[1] == '`'))
	{
		sp += 2;
		int	level = 4;
		while ((level > 1) && (sp[0] == '`'))		sp++,level--;

		new HHeading(mCT, level);
		mCT = mCT->GetNext();

		lp = sp;
		return true;
	}

	return false;
}

// AcceptLeftAlignCommand
//
// accept a left-aligned command

bool WikiParser::AcceptLeftAlignCommand(const char*& lp)
{
	if ((lp[0] != ']') && (lp[0] != '>'))	return false;

	bool	hadc = false;

	const char*	sp;

	do
	{
		if (lp[0] == ']')
		{
			// center
			if (hadc)
			{
				lp = sp;
				return true;
			}
			hadc = true;

			// if next is center, and after is textfmt or nothing, replace with new one
			// if next is not center, place new one

			if (mCT->GetNextType() != TCenter)		new HCenter(mCT);
			mCT = mCT->GetNext();
		}
		else
		{
			// blockquote
			if (mCT->GetNextType() != TBlockquote)	new HBlockquote(mCT);
			mCT = mCT->GetNext();

			hadc = false;
		}

		lp++;
		sp = lp;

		while (lp[0] == ' ')	lp++;

	} while ((lp[0] == ']') || (lp[0] == '>'));

	lp = sp;
	return true;
}

// AcceptSpaceCommand
//
// accept a command that comes after a space

bool WikiParser::AcceptSpaceCommand(const char*& lp, int minspaces)
{
	if (CS_isdigit(lp[0]) && ((lp[1] == ':') || (lp[1] == ';')))
	{
		// table command
		int	col = lp[0] - '0';
		int border = (lp[1] == ':') ? 1 : 0;

		lp += 2;

		if (mCT->GetNextType() != TTable)	new HTable(mCT, border, 8, "");
		mCT = mCT->GetNext();

		HTable*	table = (HTable*)mCT;

		if (col < table->GetCol())	table->OpenCell(col, table->GetRow() + 1, false);
		else						table->OpenCell(col, table->GetRow(), false);

		return true;
	}

	if ((lp[0] == '-') && (lp[1] == ' '))
	{
		lp += 2;

		if (mCT->GetNextType() != TVirtualList)	new HVirtualList(mCT);
		else									fprintf(mFD, "<br>\n");
		mCT = mCT->GetNext();

		return true;
	}

	int	spaces = 0;
	while (lp[0] == ' ')	lp++,spaces++;

	if (spaces < 4)
	{
		// possibly a list entry
		if ((lp[0] == '*') && (lp[1] == ' '))
		{
			lp += 2;

			while (spaces--)
			{
				if ((mCT->GetNextType() != TUnorderedList) &&
					(mCT->GetNextType() != TOrderedList))
				{
					HUnorderedList*	l = new HUnorderedList(mCT);
					l->OpenItem();
				}
				mCT = mCT->GetNext();
			}

			if (mCT->GetNextType() != TUnorderedList)	new HUnorderedList(mCT);
			mCT = mCT->GetNext();
			((HList*)mCT)->OpenItem();

			return true;
		}

		if (CS_isdigit(lp[0]))
		{
			const char*	sp = lp;

			while (CS_isdigit(lp[0]))	lp++;

			if ((lp[0] == '.') && (lp[1] == ' '))
			{
				lp += 2;

				while (spaces--)
				{
					if ((mCT->GetNextType() != TUnorderedList) &&
						(mCT->GetNextType() != TOrderedList))
					{
						HOrderedList*	l = new HOrderedList(mCT);
						l->OpenItem();
					}
					mCT = mCT->GetNext();
				}

				if (mCT->GetNextType() != TOrderedList)	new HOrderedList(mCT);
				mCT = mCT->GetNext();
				((HList*)mCT)->OpenItem();

				return true;
			}
			else
			{
				lp = sp;
			}
		}
	}

	// preformatted
	lp -= spaces;

	if (spaces < minspaces)	return false;

	lp += minspaces;

	if (mCT->GetNextType() != TPre)		new HPre(mCT);
	mCT = mCT->GetNext();

	return true;
}

// WriteLine
//
// write a line of basic text

void WikiParser::WriteLine(const char* lp)
{
	// check there's a text format item
	if (mCT->GetNextType() != TCharFmt)			new HCharFmt(mCT, mFmt);
	mCT = mCT->GetNext();

	// reset format
	mCharFmt = (HCharFmt*)mCT;
	mCharFmt->ChangeFormat(mFmt);

	// save format
	CharFmt	fmt = mFmt;

	// check for an anchor
	if (mAllowAnchor)
	{
		const char*	sp= lp;

		if (sp[0] == ' ')	sp++;
		if (sp[0] == '[')
		{
			sp++;

			char*	num = GetNumber(sp);
		
			if (num[0])
			{
				if (sp[0] == ']')
				{
					sp++;
					fprintf(mFD, "<a name=\"#%s\">[%s]</a>", num, num);
					lp = sp;
				}
			}

			delete[] num;
		}
	}

	// replace postfix commands
	char* nline = ReplacePostfix(lp);

	lp = nline;

	while (char ch = *lp)
	{
		if (CS_isalpha(ch))
		{
			AcceptAlpha(lp);
		}
		else switch (ch)
		{
		case '\'':
			AcceptQuote(lp);
			break;

		case '[':
			AcceptSquare(lp);
			break;

		case 0x01:	// strike on
			AcceptStrike(lp);
			break;

		case 0x02:	// strike off - should never occur
			lp++;
			break;

		default:
			CharOut(ch);
			lp++;
			break;
		}
	}

	delete[] nline;
	CharOut('\n');

	mFmt = fmt;
}

// ReplacePostfix
//
// replace postfix commands with prefix commands

char* WikiParser::ReplacePostfix(const char* lp)
{
	char*	nline = new char[strlen(lp) + 1];
	char*	nptr = nline;

	while (*lp)
	{
		int	count = 0;

		while ((lp[0] == '^') && (lp[1] == 'H'))
		{
			lp += 2;
			count++;
		}

		if (count)
		{
			if (nptr - nline < count)	count = nptr - nline;

			char*	tmp = new char[count];
			nptr -= count;
			memcpy(tmp, nptr, count);
			*nptr++ = 0x01;	// strike on
			memcpy(nptr, tmp, count);
			nptr += count;
			*nptr++ = 0x02;	// strike off
			delete[] tmp;
		}
		else
		{
			*nptr++ = *lp++;
		}
	}

	*nptr++ = 0;
	return nline;
}

// AcceptStrike
//
// accept a striken-out part of text

void WikiParser::AcceptStrike(const char*& lp)
{
	lp++;

	fprintf(mFD, "<strike>");
	while (*lp && (*lp != 0x02))	CharOut(*lp++);
	fprintf(mFD, "</strike>");

	lp++;
}


// AcceptPlain
//
// accept a plain part of text ?? (this has to do with {{{  }}} support)
bool WikiParser::AcceptPlain(const char*& lp) 
{
	if (!mAllowTags) 
	{
		if (lp[0] == '}' && lp[2] == '}' && lp[2] == '}' ) 
		{
			fprintf(mFD, "</pre>");
			mAllowTags = true;
			return true;
		} 
		else 
		{
			while (*lp)	CharOut(*lp++);
			fprintf(mFD, "\n");
			lp++;
			return true;
		}
	}
	return false;
}


// AcceptQuote
//
// accept part of text beginning '

void WikiParser::AcceptQuote(const char*& lp)
{
	int	quotes = 0;
	while (*lp == '\'')
	{
		quotes++;
		lp++;
	}
	if (quotes == 1)
	{
		CharOut('\'');
	}
	else
	{
		static CharFmt	_fmtmap[3][4] =
		{
			{ Italic, Plain, BoldItalic, Bold },
			{ Bold, BoldItalic, Plain, Italic },
			{ BoldItalic, Bold, Italic, Plain }
		};

		if (quotes > 4)		quotes = 4;
		
		mFmt = _fmtmap[quotes - 2][int(mFmt)];

		mCharFmt->ChangeFormat(mFmt);
	}
}

// AcceptAlpha
//
// accept part of text beginning with a letter

void WikiParser::AcceptAlpha(const char*& lp)
{
	char*	word = GetAlphaNumeric(lp);

	if (IsValidTitle(word))
	{
		OutputPageLink(word);
	}
	else if (lp[0] == ':')
	{
		AcceptURL(word, lp);
	}
	else if (!strcmp(word, "ISBN"))	// case sensitive
	{
		// ISBN 123-456-789X?
		char	rawisbn[11];

		char*	isbn = GetISBN(lp, rawisbn);

		if (isbn)
		{
			fprintf(mFD, "<a href=\"http://www.amazon.com/exec/obidos/ISBN=%s\">", rawisbn);
			fprintf(mFD, "ISBN%s</a>", isbn);
			delete[] isbn;
		}
		else
		{
			fprintf(mFD, "%s", word);
		}
	}
	else
	{
		fprintf(mFD, "%s", word);
	}

	delete[] word;
}

// AcceptSquare
//
// accept part of text beginning [

void WikiParser::AcceptSquare(const char*& lp)
{
	lp++;

	char*	num = GetNumber(lp);

	// [number] = footnote insertion
	if (num[0])
	{
		if (*lp == ']')
		{
			lp++;
			fprintf(mFD, "<a href=\"#%s\">[%s]</a>", num, num);
		}
		else
		{
			lp -= strlen(num);
			CharOut('[');
		}

		delete[] num;
		return;
	}

	delete[] num;

	// [[pagename]]
	if (*lp == '[')
	{
		lp++;

		// Search for ]]
		int len=0;

		while (lp[len] != '\0' && lp[len] != ']')
		{
			len++;
		}

		if ( lp[len] == ']' && lp[len+1] != '\0' && lp[len+1] == ']' )
		{
			// We have a possible link to a Wikipage
			char *title = new char[len+1];
			memcpy(title, lp, len);
			title[len] = '\0';

			// Make a copy of title that we can process
			char* temptitle = StringDup(title);
			
			Capitalise(temptitle);
			StripSpaces(temptitle);
			
			if (IsValidTitle(temptitle))
			{
				OutputPageLink(temptitle);
				lp += len + strlen("]]");
			}
			else
			{
				// Invalid WikiTitle, see if it is an acceptable title (alphanumeric)
				if (temptitle[0] && IsAcceptableTitle(temptitle))
				{
					OutputPageLink(temptitle);
					lp += len + strlen("]]");
				}
				else
				{
					// Non alphanumeric characters were found
					// We can't make a link of it, so just print the title
					fprintf(mFD, "[[%s", title);
					lp += len;
				}

			}

			delete[] temptitle;
			delete[] title;

			return;
		}
		else
		{
			// We could not find the ending brackets ]]
			lp--;

			CharOut('[');
			return;
		}
	}
	
	// [$command] = command
	if (*lp == '$')
	{
		lp++;

		char*	word = GetAlphaNumeric(lp);

		if (*lp != ']' && *lp != ':')  // check for ':' adds support Enhanced $commands with parameters
		{
			lp -= strlen(word) + 1;
			CharOut('[');

			delete[] word;
			return;
		}

		// Check if we are dealing with an Enhanced $command (means with parameters)
		if (*lp != ':')
		{
			// Possible normal command
			lp++;

			if (!AcceptDollarCommand(word))
			{
				lp -= strlen(word) + 2;
				CharOut('[');
			}
		}
		else
		{
			// Possible Enhanced command
			lp++;

			// Check for parameters 
			int paramlength = 0;
			for (paramlength; lp[paramlength] != '\0' && lp[paramlength] != ']'; paramlength++);

			char* params = new char[paramlength + 1];
			strncpy(params, lp, paramlength);
			params[paramlength] = '\0';

			// Make sure that the command is closed correctly before we try to accept it
			// All parameters within the tag should be on one line
			if (lp[paramlength] == ']' && AcceptEnhancedDollarCommand(word, params))
			{
				lp += paramlength + 1;
			}
			else
			{
				lp -= strlen(word) + 2;
				CharOut('[');
			}

			delete[] params;
		}

		delete[] word;
		return;
	}

	// [#anchorname] = anchor link
	// [##anchorname] = anchor definition
	if (*lp == '#')
	{
		const char*	sp = lp;
		
		sp++;

		bool	defn = false;

		if (*sp == '#')
		{
			sp++;
			defn = true;
		}

		char*	word = GetAlphaNumeric(sp);

		if (*sp == ']')
		{
			if (!defn)	fprintf(mFD, "<a href=\"#%s\">[%s]</a>", word, word);
			else		fprintf(mFD, "<a name=\"#%s\"></a>", word);

			delete[] word;

			lp = sp + 1;
			return;
		}

		delete[] word;
	}

	CharOut('[');
}

// AcceptURL
//
// accept a URL-like element which has word followed by :
// lp points to the :

void WikiParser::AcceptURL(const char* word, const char*& lp)
{
	if (!stricmp(word, "link"))			// link:Pagename
	{
		lp++;
		char* title = GetAlphaNumeric(lp);

		if (title[0])		OutputPageLink(title);
		else				fprintf(mFD, "link:");

		delete[] title;
		return;
	}

	if (!stricmp(word, "nolink") || !stricmp(word, "nl"))	// nolink:PageName or nl:PageName
	{
		lp++;
		char* title = GetAlphaNumeric(lp);

		if (title[0])		fprintf(mFD, "%s", title);
		else				fprintf(mFD, "%s:", word);
		
		delete[] title;
		return;
	}

	if (!stricmp(word, "img"))		// img:URL
	{
		lp++;
		char* title = GetURL(lp);

		if (title[0])		fprintf(mFD, "<img src=\"%s\" alt=\"Image\">", title);
		else				fprintf(mFD, "img:");

		delete[] title;
		return;
	}

	const char*	baseurl = cfg.InterWiki.Lookup(word);

	if (baseurl)
	{
		lp++;
		char* title = GetAlphaNumeric(lp);

		if (title[0])		fprintf(mFD, "<a href=\"%s%s\">%s:%s</a>", baseurl, title, word, title);
		else				fprintf(mFD, "%s:", word);

		delete[] title;
		return;
	}

	// URL
	lp -= strlen(word);		// back up to start of URL
	char* title = GetURL(lp);

	if (strlen(title) == strlen(word) + 1)
	{
		// not a URL
		fprintf(mFD, "%s:", word);

		delete[] title;
		return;
	}

	// is a URL - check protocol
	if (!strcmp(word, "local"))
	{
		// local:file.xxx is replaced with file.xxx
		if ((title[6] == '/') && (title[7] == '/'))		fprintf(mFD, "<a href=\"%s\">%s</a>", title + 8, title + 8);
		else											fprintf(mFD, "<a href=\"%s\">%s</a>", title + 6, title + 6);
		
		return;
	}

	if (!strcmp(word, "ilocal"))
	{
		// ilocal://file.xxx is replaced with inline image
		if ((title[7] == '/') && (title[8] == '/'))		fprintf(mFD, "<img src=\"%s\" alt=\"Image\">", title + 9);
		else											fprintf(mFD, "<img src=\"%s\" alt=\"Image\">", title + 7);

		return;
	}

	if (!strcmp(word, "mailto"))
	{
		fprintf(mFD, "<a href=\"%s\">%s</a>", title, title + 7);
		return;
	}

	if (!strcmp(word, "http") ||
		!strcmp(word, "https") ||  
		!strcmp(word, "file") ||  
		!strcmp(word, "ftp"))
	{
		fprintf(mFD, "<a href=\"%s\">%s</a>", title, title);
		return;
	}

	lp -= strlen(title);
	lp += strlen(word) + 1;
	fprintf(mFD, "%s:", word);
}

// AcceptFormCommand
//
// handle a [?<command>] command

bool WikiParser::AcceptFormCommand(const char* word)
{
	if (!stricmp(word, "titlesearch"))
	{
		// title search
		mCT->SetSubtag(0);
		WriteSearchForm("title_search", "", mFD);
		return true;
	}

	if (!stricmp(word, "fullsearch"))
	{
		// full search
		mCT->SetSubtag(0);
		WriteSearchForm("full_search", "", mFD);
		return true;
	}
	
	if (!stricmp(word, "googlesearch"))
	{
		// Google search
		mCT->SetSubtag(0);
		WriteGoogleSearchForm("", mFD);
		return true;
	}
	
	if (!stricmp(word, "amazonsearch"))
	{
		// Amazon search
		mCT->SetSubtag(0);
		WriteAmazonSearchForm("", mFD);
		return true;
	}

	return false;
}

// AcceptDollarCommand
//
// handle a [$<command>] command

bool WikiParser::AcceptDollarCommand(const char* word)
{
	if (!stricmp(word, "pagetime"))
	{
		fprintf(mFD, "%s", mPage ? mPage->GetTimeString() : "??:??");
		return true;
	}

	if (!stricmp(word, "pagedate"))
	{
		fprintf(mFD, "%s", mPage ? mPage->GetDateString() : "??/??/????");
		return true;
	}

	if (!stricmp(word, "defpage"))
	{
		OutputPageLink(cfg.DefaultPage);
		return true;
	}

	if (!stricmp(word, "username"))
	{
		fprintf(mFD, "%s", mUser->GetUsername());
		return true;
	}

	if (!stricmp(word, "homepage"))
	{
		OutputPageLink(mUser->GetHomepage());
		return true;
	}

	if (!stricmp(word, "userperm"))
	{
		fprintf(mFD, "%s", mUser->GetPermString());
		return true;
	}

	if (!stricmp(word, "version"))
	{
		fprintf(mFD, "%s", WIKISERVER_VERSION);
		return true;
	}

	if (!stricmp(word, "this"))
	{
		if (mPage)	OutputPageLink(mPage->GetTitle());
		else		fprintf(mFD, "???");
		return true;
	}

	if (!stricmp(word, "reverselink"))
	{
		if (mPage)
		{
			cfg.Pages->ReverseLinkList(mPage, mFD, mUser, mConvert);
		}
		else
		{
			fprintf(mFD, "???");
		}
		return true;
	}

	return false;
}

// AcceptEnhancedDollarCommand
//
// handle a [$<command>:<params>] command (a.k.a. Enhanced $commands)

bool WikiParser::AcceptEnhancedDollarCommand(const char* word, const char* params)
{
	if (!stricmp(word, "html"))
	{
		int brackets = 0;
		int singlequotes = 0;
		int doublequotes = 0;
		bool intag = false;


		// Sanitycheck on params to see if brackets and quotes match
		for (int i = 0; i < (int)strlen(params); i++)
		{
			if ( intag && params[i] == '\'')	{ singlequotes++; }
			if ( intag && params[i] == '\"')	{ doublequotes++; }

			if (params[i] == '<')	
			{ 
				brackets++; 
				intag = true;
			}

			if ( intag && params[i] == '>')	
			{ 
				// Quotes should be closed within the HTML tag
				if (singlequotes & 1 || doublequotes & 1)
				{
					// Do not print the error for now, not sure if it is desired.
					//	fprintf(mFD, "The HTML code has not been processed because the number of quotes within a tag does not match.<br>");
					return false;
				}

				brackets--; 
				intag = false;
			}
		}

		// Abort if the number of < and > is not the same
		if (brackets != 0)
		{

			// Do not print the error for now, not sure if it is desired.
			//	fprintf(mFD, "The HTML code has not been processed because the number of brackets does not match.<br>");
			return false;
		}

		// Just print the params (=HTML code)
		fprintf(mFD, "%s", params);

		return true;
	}

	if (!stricmp(word, "edit"))
	{
		// Create a link to edit the page mentioned in the params
		if (!mConvert)
		{
			fprintf(mFD, "<a href=\"EditPage?page=%s\">Edit %s</a>", params, params);
		}
		else
		{
			fprintf(mFD, "Edit %s", params);
		}

		return true;
	}

	if (!stricmp(word, "reverselink"))
	{
		// Create a bulleted list of all pages that link to the page provided in the params
		Page* tempPage;

        tempPage = cfg.Pages->FindPage(params);

		if (tempPage)
		{
			cfg.Pages->ReverseLinkList(tempPage, mFD, mUser, mConvert);
		}

		return true;
	}


	return false;
}

// AcceptFileInsertion
//
// handle a [!<file>] command

void WikiParser::AcceptFileInsertion(const char* word)
{	
	if (IsAcceptableTitle(word))
	{
		Page*	pp = cfg.Pages->FindPage(word);

		if (!pp)
		{
			fprintf(mFD, "%s<a href=\"%s\">?</a>", word, word);
		}
		else if (pp->IsLoaded() || (pp->GetFlags() & PF_META) || !cfg.Pages->LoadPage(pp))
		{
			// loaded already (cannot have recursion) or meta - don't include
			fprintf(mFD, "(Can't include ");
			OutputPageLink(word);
			CharOut(')');
		}
		else
		{
			WikiParser	parser;

			parser.WikiToHTML(mFD, pp->GetData(), mFmt, mPage, mUser, mConvert);

			cfg.Pages->UnloadPage(pp);
		}
	}
	else
	{
		char	fn[_MAX_PATH];
		sprintf(fn, "%s\\%s", cfg.Pathname, word);

		FILE*	fd = fopen(fn, "rb");

		if (!fd)
		{
			fprintf(mFD, "%s<a href=\"/UploadFile?file=%s\">&gt;</a>", word, word);
		}
		else
		{
			fclose(fd);
			fprintf(mFD, "<pre>%s</pre>", word);
		}
	}
}
