// WikiServer
//
// Wiki server class

#include "Inc.h"

// WikiServer
//
// create PageSet
// create metapages
// load userlist
// load configuration

WikiServer::WikiServer(int port, FILE* logfd, const char* rootdir) : Server(port, logfd)
{
	// init other parts of cfg
	cfg.Pages = 0;
	cfg.Pathname = StringDup(rootdir);

	// create page set
	mPageSet = new PageSet();
	cfg.Pages = mPageSet;

	// create meta-pages
	mEditPage = mPageSet->CreatePage("EditPage", PF_META | PF_PAGEPARM);
	mDumpAllPages = mPageSet->CreatePage("DumpAllPages", PF_META);
	mListAllPages = mPageSet->CreatePage("ListAllPages", PF_META);
	mListOrphans = mPageSet->CreatePage("ListOrphans", PF_META);
	mRecentChanges = mPageSet->CreatePage("RecentChanges", PF_META);
	mReverseLink = mPageSet->CreatePage("ReverseLink", PF_META | PF_PAGEPARM);
	mAdminLogin = mPageSet->CreatePage("AdminLogin", PF_META);
	mAuthorLogin = mPageSet->CreatePage("AuthorLogin", PF_META);
}

// ~WikiServer
//
// delete pageset

WikiServer::~WikiServer()
{
	delete mPageSet;
	delete[] cfg.Pathname;
}

// GetServerName
//
// return this server's name

const char* WikiServer::GetServerName()
{
	return "Eddie's WikiServer V1.1";
}

// Get
//
// attempt to satisfy the HTTP request

int WikiServer::Get(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	// get authorization
	const char* auth = rq->GetAuth();
	
	if (!auth)	auth = ":";

	const UserInfo*	user = cfg.QueryAuth(auth);

	// if no anonymous user return 401
	if (!user)				return 401;

	// if user has no read permissions return 401
	if (!user->Auth('r'))	return 401;

	// get URI; set the default for /
	const char*	uri = rq->GetURI();
	if (uri[0] == '/')	uri++;
	if (!uri[0])
	{
		if (strcmp(auth, ":") && user->Auth('a'))
		{
			// send a named admin to AdminLogin, whatever RootPage says
			uri = "AdminLogin";
		}
		else
		{
			uri = cfg.RootPage;
		}
	}

	//
	// check for POST operations
	//

	// page?
	const Header*	pagep = rq->GetParms()->Find("page");

	if (pagep && !strcmp(uri, "SubmitPage"))
	{
		hdrs.AddHeader("Content-Type", cfg.ContentType);
		return SubmitPage(pagep->mData, file, rq, hdrs, user);
	}

	// title_search?
	const Header*	titlep = rq->GetParms()->Find("title_search");

	if (titlep)
	{
		hdrs.AddHeader("Content-Type", cfg.ContentType);
		return TitleSearch(titlep->mData, file, rq, hdrs, user);
	}

	// full_search?
	const Header*	fullp = rq->GetParms()->Find("full_search");

	if (fullp)
	{
		hdrs.AddHeader("Content-Type", cfg.ContentType);
		return FullSearch(fullp->mData, file, rq, hdrs, user);
	}

	//
	// no operation: it's a GET
	//

	// look for an extension (look from back to front of uri to allow for paths with dots in it)
    const char* uptr = uri + strlen(uri);
	while ((uptr != uri) && (*uptr != '/') && (*uptr != '.')) uptr--;

	if (*uptr == '.')  // There is an extension
	{
		return GetRaw(file, uri, uptr + 1, hdrs, user);
	}
	else
	{
		return ServeWikiName(uri, file, rq, hdrs, user);
	}
}

// ServeWikiName
//
// when we don't have a special case, this serves the page

int WikiServer::ServeWikiName(const char* uri, TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	Page*	pp = mPageSet->FindPage(uri);

	if (pp)
	{
		// page exists
		if (pp->GetFlags() & PF_META)
		{
			// meta-page - use a Cmd handler
			return Cmd(pp, file, rq, hdrs, user);
		}

		// open file
		FILE*	fd = file.Open("w");
		if (!fd)	return 500;

		// output page
		hdrs.AddHeader("Content-Type", cfg.ContentType);
		return mPageSet->WriteAsHTML(pp, fd, user) ? 200 : 500;
	}

	// page does not exist
	if (IsValidTitle(uri) || IsAcceptableTitle(uri))
	{
		// forge an EditPage?page=<page>
		rq->GetParms()->AddHeader("page", uri);
		hdrs.AddHeader("Content-Type", cfg.ContentType);
		return EditPage(file, rq, hdrs, user);
	}

	return 404;
}

// GetRaw
//
// return a raw file

int WikiServer::GetRaw(TempFile& file, const char* filename, const char* ext, Headers& hdrs, const UserInfo* user)
{
	// disallow filenames with \ or / in them for security
	if (strchr(filename, '\\'))	return 403;

	// MvdW allow for / so we can organise pictures, docs etc. in folders
	//	if (strchr(filename, '/'))	return 403;

	// Don't do files with .. for security
	if (strstr(filename, ".."))	return 403;

	// disallow .wkc files for security
	if (!stricmp(ext, "wkc"))	return 403;

	// make filename
	char*	fname = new char[strlen(cfg.Pathname) + strlen(filename) + 2];
	strcpy(fname, cfg.Pathname);
	strcat(fname, "\\");
	strcat(fname, filename);

	// open
	FILE*	ifd = fopen(fname, "rb");
	if (!ifd)
	{
		delete[] fname;
		return 404;
	}
	fclose(ifd);

	// set clone file
	file.CloneExistingFile(fname);
	delete[] fname;

	// add correct Content-Type
	const char*	type = cfg.GetMimeType(ext);

	hdrs.AddHeader("Content-Type", type);

	return 200;
}

// SubmitPage
//
// form handler for edit submission

int WikiServer::SubmitPage(const char* pagename, TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	// check write perm
	if (!user->Auth('w'))			return 401;

	// get other parameters
	const	Header*	textp = rq->GetParms()->Find("text");
	if (!textp)	return 400;

	const	Header*	seqnop = rq->GetParms()->Find("seqno");

	// define the page
	Page*	pp = mPageSet->CreatePage(pagename);
	if (!pp)
	{
		pp = mPageSet->FindSysPage(pagename);
		if (!pp)	return 400;
	}

	if (pp->GetFlags() & PF_META)	return 400;
	if (pp->GetFlags() & PF_ADMINEDIT)
	{
		if (!user->Auth('a'))		return 401;
	}

	if (seqnop)
	{
		int	seqno = atoi(seqnop->mData);

		if (seqno < pp->GetSeqNo())
		{
			// earlier version - show merge page
			FILE*	fd = file.Open("w");
			if (!fd)	return 500;

			return mPageSet->WriteMergePage(pp, textp->mData, fd, user) ? 200 : 500;
		}
	}

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	if (!textp->mData[0] && user->Auth('a') && !(pp->GetFlags() & PF_NONWIKI))
	{
		mPageSet->DeletePage(pp);

		return mPageSet->WriteDeletedMessage(pagename, fd, user) ? 200 : 500;
	}

	mPageSet->EditPage(pp, textp->mData, strlen(textp->mData));

//	return mPageSet->WriteEditedMessage(pp, fd, user) ? 200 : 500; // This writes the thankyou message.
	return ServeWikiName(pagename, file, rq, hdrs, user);
}

// TitleSearch
//
// form handler for a title search

int WikiServer::TitleSearch(const char* string, TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	mPageSet->TitleSearch(string, fd, user);

	return 200;
}

// FullSearch
//
// form handler for a full search

int WikiServer::FullSearch(const char* string, TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	mPageSet->FullSearch(string, fd, user);

	return 200;
}
