// WikiServerCmd
//
// Wiki server commands
//
// the WikiServer functions validate and extract parameters etc. then
// call into PageSet functions [PageHTML.cpp] to output the page itself

#include "Inc.h"

// Cmd
//
// route command to handler

int WikiServer::Cmd(Page* page, TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	if (page == mEditPage)		return EditPage(file, rq, hdrs, user);
	if (page == mDumpAllPages)	return DumpAllPages(file, rq, hdrs, user);
	if (page == mListAllPages)	return ListAllPages(file, rq, hdrs, user);
	if (page == mListOrphans)	return ListOrphans(file, rq, hdrs, user);
	if (page == mRecentChanges)	return RecentChanges(file, rq, hdrs, user);
	if (page == mReverseLink)	return ReverseLink(file, rq, hdrs, user);
	if (page == mAdminLogin)	return AdminLogin(file, rq, hdrs, user);
	if (page == mAuthorLogin)	return AuthorLogin(file, rq, hdrs, user);

	return 404;
}

// EditPage
//
// page = page name
// rows = # rows
// cols = #cols

int WikiServer::EditPage(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	if (!user->Auth('w'))	return 401;

	const Header*	value = rq->GetParms()->Find("page");
	if (!value)	return 400;

	const Header*	rows = rq->GetParms()->Find("rows");
	const Header*	cols = rq->GetParms()->Find("cols");

	int	nrows = rows ? atoi(rows->mData) : 0;
	int	ncols = cols ? atoi(cols->mData) : 0;

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	// output page
	hdrs.AddHeader("Content-Type", cfg.ContentType);
	return mPageSet->WriteEditPage(value->mData, nrows, ncols, fd, user);
}

// DumpAllPages
//
// no parms

int WikiServer::DumpAllPages(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	if (!user->Auth('a'))	return 401;

	FILE*	fd = file.Open("wb");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", "application/wikiserver");
	return mPageSet->DumpAllPages(fd, user);
}

// ListAllPages
//
// no parms

int WikiServer::ListAllPages(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);
	mPageSet->ListAllPages(fd, user);

	return 200;
}

// ListOrphans
//
// no parms

int WikiServer::ListOrphans(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);
	mPageSet->ListOrphans(fd, user);

	return 200;
}

// RecentChanges
//
// no parms

int WikiServer::RecentChanges(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);
	mPageSet->RecentChanges(fd, user);

	return 200;
}

// ReverseLink
//
// page = page name

int WikiServer::ReverseLink(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	const Header*	value = rq->GetParms()->Find("page");
	if (!value)	return 400;

	Page*	page = mPageSet->FindPage(value->mData);
	if (!page)	return 404;

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);
	mPageSet->ReverseLink(page, fd, user);

	return 200;
}

// AdminLogin
//
// cmd = command to run, else give menu

int WikiServer::AdminLogin(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	if (!user->Auth('a'))	return 401;

	const Header*	value = rq->GetParms()->Find("cmd");

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);

	if (!value)
	{
		mPageSet->AdminLogin(fd, user);
	}
	else
	{
		if (!stricmp(value->mData, "shutdown"))
		{
			ShutdownAfterTransaction();

			fprintf(fd, "<html><head><title>The server has shut down</title></head>\n");
			fprintf(fd, "<body><h1>The server has shut down</h1><hr></body></html>\n");
		}
		else if (!stricmp(value->mData, "convert"))
		{
			mPageSet->ConvertPages(fd, user);
		}
		else
		{
			mPageSet->AdminLogin(fd, user);
		}
	}

	return 200;
}

// AuthorLogin
//
// this page serves the main purpose of requiring write permission to VIEW it - so
// anyone browsing to AuthorLogin without write permission is presented with a login dialog
//
// the page forwards to the current user's homepage

int WikiServer::AuthorLogin(TempFile& file, HTTP_RQ* rq, Headers& hdrs, const UserInfo* user)
{
	if (!user->Auth('w'))		return 401;
	if (user->Auth('a'))		return 401;		// AuthorLogin also allows you to log *out* from being admin
	if (user->IsAnonymous())	return 401;

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", cfg.ContentType);

	Page*	pp = 0;

	if (stricmp(user->GetHomepage(), "AuthorLogin"))
	{
		// go to author's homepage
		return ServeWikiName(user->GetHomepage(), file, rq, hdrs, user);
	}
	
	return ServeWikiName(cfg.DefaultPage, file, rq, hdrs, user);
}
