// headers.h
//
// header set

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "headers.h"
#include "utility.h"
#include "os.h"

Headers::Headers()
{
	mParms = NULL;
}

Headers::~Headers()
{
	DeleteAll();
}

void Headers::DeleteAll()
{
	Header*	hp = mParms;

	while (hp)
	{
		Header*	next = hp->mNext;
		delete[] hp->mName;
		delete hp;
		hp = next;
	}

	mParms = 0;
}

void Headers::AddHeader(const char* name, const char* data)
{
	Header*	hp = new Header;

	hp->mName = new char[strlen(name) + 1];
	strcpy(hp->mName, name);

	hp->mData = new char[strlen(data) + 1];
	strcpy(hp->mData, data);

	hp->mNext = mParms;
	mParms = hp;
}

void Headers::AddFromHTTP(const char* httprq)
{
	if (!httprq)	return;

	char*	hdr = StringDup(httprq);

	for (;;)
	{
		char*	key;

		// split at <LF>
		hdr = StringSplit(hdr, 0x0A, key);

		// no more keys?
		if (!key || !key[0])	break;

		// scan for colon
		char* hptr = key;
		while (*hptr && (*hptr != ':'))	hptr++;

		// add header
		if (*hptr)
		{
			*hptr++ = '\0';
			while (*hptr == ' ')	hptr++;
			AddHeader(key, hptr);
		}
		else
		{
			AddHeader(key, "");
		}

		delete[] key;
	}

	delete[] hdr;
}

void Headers::AddFromCGI(const char* cgistr)
{
	if (!cgistr)	return;

	char*	data = StringDup(cgistr);

	while (data)
	{
		char*	line;

		// split at &, else take remainder
		data = StringSplit(data, '&', line);
		if (!line)
		{
			line = data;
			data = NULL;
		}

		// split at =
		char*	key;

		line = StringSplit(line, '=', key);
		if (key)
		{
			// parse first
			URLDecode(line);
//			UTFDecode(line);
			AddHeader(key, line);
		}
		else if (line[0])
		{
			AddHeader(line, "");
		}

		delete[] line;
		delete[] key;
	}
}

void Headers::AddFromCookie(const char* cookie)
{
	if (!cookie)	return;

	char*	data = StringDup(cookie);

	while (data)
	{
		char*	line;

		// split at ;, else take remainder
		data = StringSplit(data, ';', line);
		if (!line)
		{
			line = data;
			data = NULL;
		}

		// split at =
		char*	key;

		line = StringSplit(line, '=', key);
		if (key)	AddHeader(key, line);

		delete[] line;
		delete[] key;
	}
}

bool Headers::AddFromFile(const char* file)
{
	char*	filedata = LoadFile(file);

	bool	rc = AddFromFileData(filedata);

	delete[] filedata;

	return rc;
}

bool Headers::AddFromFileData(const char* filedata)
{
	const char*	fptr = filedata;

	while (fptr[0])
	{
		char* line = GetLine(fptr);

		if (line[0] != ';')	// comments are allowed in files
		{
			// split at =
			char* key;

			line = StringSplit(line, '=', key);
			if (key)	AddHeader(key, line);

			delete[] key;
		}

		delete[] line;
	}

	return true;
}

bool Headers::WriteToFile(const char* file) const
{
	FILE*	fd = fopen(file, "w");

	if (!fd)	return false;

	for (Header* hp = mParms; hp; hp = hp->mNext)
	{
		fprintf(fd, "%s=%s\n", hp->mName, hp->mData);
	}

	fclose(fd);

	return true;
}

const Header* Headers::Find(const char* name, int which) const
{
	for (Header* hp = mParms; hp; hp = hp->mNext)
	{
		if (!stricmp(hp->mName, name))
		{
			if (!which--)	return hp;
		}
	}

	return 0;
}

const char* Headers::Lookup(const char* name, const char* def) const
{
	const Header*	hdr = Find(name);

	if (!hdr)	return def;

	return hdr->mData;
}

void Headers::Write(FILE* fd) const
{
	for (Header* hp = mParms; hp; hp = hp->mNext)
	{
		fprintf(fd, "%s: %s\n", hp->mName, hp->mData);
	}
}

int Headers::Count() const
{
	int	count = 0;

	for (Header* hp = mParms; hp; hp = hp->mNext)	count++;

	return count;
}

const Header* Headers::Index(int which) const
{
	for (Header* hp = mParms; hp; hp = hp->mNext)
	{
		if (!which--)	return hp;
	}

	return 0;
}

int Headers::FindIndex(const char* name, int which) const
{
	int	ix = 0;

	for (Header* hp = mParms; hp; hp = hp->mNext)
	{
		if (!stricmp(hp->mName, name))
		{
			if (!which--)	return ix;
		}
		ix++;
	}

	return -1;
}
