// os_win32
//
// Win32 OS-specific routines

#ifdef WIN32

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include <windows.h>

#include <sys/utime.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "os.h"
#include "utility.h"

// OS_Init
//
// do any OS-specific initialization

void OS_Init()
{
	// disable retry dialog
	SetProcessShutdownParameters(0x100, SHUTDOWN_NORETRY);
}

// OS_DetachConsole
//
// detach from console window if possible

void OS_DetachConsole()
{
	// close console window
	FreeConsole();
}

// OS_InitSocketAPI
//
// initialize socket API

bool OS_InitSocketAPI()
{
	// begin Windows sockets
	WORD	wVersionRequested = MAKEWORD(2, 0);
	WSADATA	wsaData;

	if (WSAStartup(wVersionRequested, &wsaData))
	{
		return false;
	}

	return true;
}

// OS_TermSocketAPI
//
// terminate socket API

void OS_TermSocketAPI()
{
	WSACleanup();
}

// OS_GetSocketError
//
// get socket error code

int OS_GetSocketError()
{
	return WSAGetLastError();
}

// OS_MakeTempFilename
//
// make a temp filename using the given tag

void OS_MakeTempFilename(char* buffer, const char* tag)
{
	char	tp[MAX_PATH];
	GetTempPath(MAX_PATH, tp);

	sprintf(buffer, "%s\\%s_%d.txt", tp, tag, GetCurrentThreadId());
}

// OS_Yield
//
// yield for a short time

void OS_Yield()
{
	Sleep(0);
}

// OS_GetModifiedTime
//
// get the modified time for a file

time_t OS_GetModifiedTime(const char* pathname)
{
	struct _stat	stat;

	_stat(pathname, &stat);

	return stat.st_mtime;
}

// OS_SetModifiedTime
//
// set the modified time for a file

void OS_SetModifiedTime(const char* pathname, time_t time)
{
	_utimbuf	tbuf;

	tbuf.actime = time;
	tbuf.modtime = time;

	_utime(pathname, &tbuf);
}

// OS_ScanDir
//
// given a directory, calls the callback for every file
// in the directory

void OS_ScanDir(const char* dirname, void (*cb)(const char* dirname, const char* filename, const char* basename, const char* ext, void* apikey), void* apikey)
{
	char			fn[_MAX_PATH];
	WIN32_FIND_DATA	fdata;

	sprintf(fn, "%s\\*.*", dirname);

	HANDLE hnd = FindFirstFile(fn, &fdata);

	if (hnd != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)	continue;

			// parse the filename
			char	basename[_MAX_PATH];
			char	ext[_MAX_PATH];

			char*	s = fdata.cFileName;
			char*	e = s + strlen(s);

			while (e > s)
			{
				if (e[-1] == '.')	break;
				e--;
			}

			if (e > s)
			{
				e--;
				memcpy(basename, s, e - s);
				basename[e - s] = 0;
				
				strcpy(ext, e + 1);
			}
			else
			{
				strcpy(basename, s);
				ext[0] = 0;
			}

			// callback
			cb(dirname, s, basename, ext, apikey);

		} while (FindNextFile(hnd, &fdata));
	}
}

// GetTimeSys
//
// get system time for time_t

static void GetTimeSys(struct tm* stime, time_t t)
{
	struct tm*	rc = localtime(&t);

	if (rc)		memcpy(stime, rc, sizeof(struct tm));
	else		memset(stime, 0, sizeof(struct tm));
}

// OS_GetDateString
//
// convert a time_t to a (static) date string

const char* OS_GetDateString(time_t t)
{
	static char	date[64];

	struct tm	stime;

	GetTimeSys(&stime, t);

	sprintf(date, "%s %d, %d", MonthName(stime.tm_mon + 1), stime.tm_mday, stime.tm_year + 1900);

	return date;
}

// OS_GetTimeString
//
// convert a time_t to a (static) time string

const char* OS_GetTimeString(time_t t)
{
	static char	time[128];

	struct tm	stime;

	GetTimeSys(&stime, t);

	sprintf(time, "%02d:%02d", stime.tm_hour, stime.tm_min);

	return time;
}

// OS_IsSameDay
//
// see if two time_ts are on the same day

bool OS_IsSameDay(time_t t1, time_t t2)
{
	struct tm	st1, st2;

	GetTimeSys(&st1, t1);
	GetTimeSys(&st2, t2);

	if (st1.tm_year != st2.tm_year)		return false;
	if (st1.tm_mon != st2.tm_mon)		return false;
	if (st1.tm_mday != st2.tm_mday)		return false;

	return true;
}

#endif	// WIN32
