// server.cpp
//
// HTTP server

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "server.h"
#include "utility.h"
#include "os.h"
#include "os_sockets.h"

Server::Server(int port, FILE* logfd) : mPort(port), mLog(logfd)
{
	Log("%s,%s,,INIT,/,0,000,000 Server reset\n", GetSysDate(), GetSysTime());

	mAlive = true;
}

Server::~Server()
{
}

// Log format:
//
// <date>
// <time>
// <ip address>
// <method>		(get or post)
// <uri>
// <number of bytes sent excluding headers>
// <result code>
// <result string>

void Server::Log(char* fmt, ...)
{
	if (!mLog)	return;

	static char	buffer[4096];

	va_list	args;
	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);

	fputs(buffer, mLog);
	fflush(mLog);
}

void Server::FatalError(char* fmt, ...)
{
	if (!mLog)	return;

	static char	buffer[4096];

	va_list	args;
	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);

	fprintf(mLog, "%s,%s,,ERROR,/,0,999,999 %s\n", GetSysDate(), GetSysTime(), buffer);
	fflush(mLog);
}

void Server::AddServerHeaders(Headers& hdrs)
{
	char	buffer[1024];

	sprintf(buffer, "Eddie's MiniWebServer %s running %s", WIKISERVER_VERSION, GetServerName());

	hdrs.AddHeader("Server", buffer);
	hdrs.AddHeader("Pragma", "no-cache");
}

bool Server::Run()
{
	// init server files
	CreateErrorPages();

	// init socket API
	if (!OS_InitSocketAPI())
	{
		FatalError("Cannot initialize socket API");
		return false;
	}

	// get socket
	SOCKET	s;
	int		rc = GetPort(s);

	if (rc == -1)
	{
		FatalError("Cannot open socket");
		return false;
	}
	else if (rc == -2)
	{
		FatalError("Address (127.0.0.1:%d) in use", mPort);
		return false;
	}

	Log("%s,%s,,INIT,/,0,001,001 Server init complete on port %d\n", GetSysDate(), GetSysTime(), mPort);
	Log("%s,%s,,INIT,/,0,002,002 WikiServer version %s ready\n", GetSysDate(), GetSysTime(), WIKISERVER_VERSION);

	// serve
	while (mAlive)
	{
		struct sockaddr raddr;
		struct sockaddr_in *rap = (struct sockaddr_in *)&raddr;

		AcceptLengthType	raddrlen = AcceptLengthType(sizeof(raddr));

		SOCKET as;

		// accept connection
		for (;;)
		{
			as = accept(s, &raddr, &raddrlen);
			if (as != INVALID_SOCKET)						break;		// connected
			if (OS_GetSocketError() == SERR_EINTR)			continue;	// harmless - try again
			if (OS_GetSocketError() != SERR_EWOULDBLOCK)	break;		// actual error
			OS_Yield();
		}

		// begin CSV log entry
		Log("%s,%s", GetSysDate(), GetSysTime());

		if (as == INVALID_SOCKET)
		{
			Log(",,ERROR,/,0,999, 999 Accept failed\n");
		}
		else
		{
			Log(",%s", inet_ntoa(rap->sin_addr));

			// set new socket to non-blocking
			u_long	one = 1;
			if (ioctlsocket(as, FIONBIO, &one) != SOCKET_ERROR)
			{
				// read message from socket
				HTTP_RQ* rq = ReadMessage(as);

				if (rq)
				{
					if (!rq->GetReq())
					{
						Log(",ERROR,/");
						HTTPError(as, 400);
					}
					else
					{
						RunCmd(as, rq);
					}

					delete rq;
				}
				else
				{
					Log(",ERROR,/");
					HTTPError(as, 500);
				}
			}

			closesocket(as);
			Log("\n");
		}
	}

	OS_TermSocketAPI();

	return true;
}

int Server::GetPort(SOCKET& s)
{
	// create socket
	struct sockaddr_in inaddr;

	memset(&inaddr, 0, sizeof(inaddr));

	inaddr.sin_family		= AF_INET;
	inaddr.sin_addr.s_addr	= INADDR_ANY;
	inaddr.sin_port			= htons(mPort);

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == INVALID_SOCKET)	return -1;

	// bind socket
	if (bind(s, (struct sockaddr *)&inaddr, sizeof(inaddr)) == SOCKET_ERROR)
	{
		closesocket(s);
		if (OS_GetSocketError() == SERR_EADDRINUSE)	return -2;
		else										return -1;
	}

	// listen
	while (listen(s, 5) == SOCKET_ERROR)
	{
		if (OS_GetSocketError() != SERR_EINTR)
		{
			closesocket(s);
			return -1;
		}
	}

	// set to blocking mode (should be the default)
	u_long zero = 0;
	ioctlsocket(s, FIONBIO, &zero);

	return 0;
}

static int Send(SOCKET s, const char* dbuf, int dlen, int flags)
{
	int olen = dlen;
	int	slen;

	for (;;)
	{
		if (dlen < 65536)
		{
			slen = send(s, dbuf, dlen, flags);
			if (slen != SOCKET_ERROR)
			{
				if (slen == dlen)	return olen;
				dbuf += slen;
				dlen -= slen;
			}
			else
			{
				int	err = OS_GetSocketError();

				if (err != SERR_EWOULDBLOCK)	break;	// proper error
			}
			OS_Yield();
		}
		else
		{
			slen = send(s, dbuf, 65536, flags);
			if (slen != SOCKET_ERROR)
			{
				dbuf += 65536;
				dlen -= 65536;
			}
			else
			{
				int	err = OS_GetSocketError();

				if (err != SERR_EWOULDBLOCK)	break;	// proper error
			}
			OS_Yield();
		}
	}

	return SOCKET_ERROR;
}

static int Recv(SOCKET s, char* buf, int dlen, int flags)
{
	int	rlen;

	for (;;)
	{
		rlen = recv(s, buf, dlen, 0);
		if (!rlen)										return SOCKET_ERROR;	// connection is closed
		if (rlen != SOCKET_ERROR)						return rlen;
		if (OS_GetSocketError() != SERR_EWOULDBLOCK)	break;					// real error
		OS_Yield();
	}

	return SOCKET_ERROR;
}

// CreateErrorPages
//
// create pages used by HTTPError function

void Server::CreateErrorPages()
{
	CreateErrorPage(400, "Bad request", "The server cannot complete your request");
	CreateErrorPage(401, "Unauthorized", "You are not authorized to view that page");
	CreateErrorPage(403, "Forbidden", "You are forbidden to view that page");
	CreateErrorPage(404, "Not found", "The page you requested was not found");

	CreateErrorPage(500, "Internal error", "The server has suffered an internal error");
	CreateErrorPage(501, "Not implemented", "The server does not implement that request");
}

void Server::CreateErrorPage(int status, char* desc, char* verbiage)
{
	char	temp[_MAX_PATH];
	sprintf(temp, "Error%d.html", status);
	FILE*	fd = fopen(temp, "rb");
	if (!fd)
	{
		fd = fopen(temp, "w");

		if (fd)
		{
			fprintf(fd, "%s\n", StatusLine(status));

			Headers	hdr;

			hdr.AddHeader("Content-Type", "text/html");
			AddServerHeaders(hdr);

			if (status == 401)
			{
				const char*	realm = GetServerName();
				char*		realmhdr = new char[strlen(realm) + 32];

				sprintf(realmhdr, "Basic realm=\"%s\"", realm);

				hdr.AddHeader("WWW-Authenticate", realmhdr);
				hdr.AddHeader("Expires", "Thu, 13 Dec 1969, 10:29:00 GMT");

				delete[] realmhdr;
			}

			hdr.Write(fd);

			fprintf(fd, "\n<html><head><title>\n%d - %s\n", status, desc);
			fprintf(fd, "</title></head>\n<body><h1>\n%d - %s\n</h1>\n", status, desc);
			fprintf(fd, "<p>%s</p>\n</body></html>\n", verbiage);

			fclose(fd);
		}
	}
	else
	{
		fclose(fd);
	}
}

void Server::HTTPError(SOCKET s, int status)
{
	const char* err = StatusLine(status);

	// look for response file
	size_t dlen;

	char temp[_MAX_PATH];
	sprintf(temp, "Error%d.html", status);
	char*	dbuf = LoadFile(temp);
	if (dbuf)
	{
		dlen = strlen(dbuf);
		Send(s, dbuf, dlen, 0);
		delete[] dbuf;
	}
	else
	{
		sprintf(temp, "%s\x0D\x0A\x0D\x0A", err);
		dlen = strlen(temp);
		Send(s, temp, dlen, 0);
	}

	Log(",%d,%d,%s", dlen, status, err);
}

// http_terminator
//
// looks for [<0D>]<0A>[<0D>]<0A> in a block
//
// returns position of next character, else 0

static int http_terminator(char* mbuf, int len)
{
	int	conseq = 0;

	for (int ii = 0; ii < len; ii++)
	{
		if (mbuf[ii] == 0x0A)
		{
			conseq++;
			if (conseq == 2)	return (ii + 1);
		}
		else if (mbuf[ii] == 0x0D)
		{
			// ignore
		}
		else
		{
			conseq = 0;
		}
	}

	return 0;
}

// ReadMessage
//
// read a message

HTTP_RQ* Server::ReadMessage(SOCKET s)
{
	char	buf[512];		// buffer for data
	char*	mbuf = NULL;	// cumulative buffer
	size_t	len = 0;		// length of mbuf[]
	size_t	tpos = 0;		// end of header

	for (;;)
	{
		int	rlen = Recv(s, buf, 512, 0);

		if (rlen == SOCKET_ERROR)
		{
			delete[] mbuf;
			return NULL;
		}

		if (len)
		{
			// re-alloc
			char*	nbuf = new char[len + rlen + 1];
			memcpy(nbuf, mbuf, len);
			memcpy(nbuf + len, buf, rlen);
			delete[] mbuf;
			mbuf = nbuf;
			len += rlen;

			tpos = http_terminator(mbuf, len);

			if (tpos)	break;
		}
		else
		{
			mbuf = new char[rlen + 1];
			memcpy(mbuf, buf, rlen);
			len = rlen;

			tpos = http_terminator(mbuf, len);

			if (tpos)	break;
		}
	}

	// copy header out
	char*	hdr = new char[tpos + 1];
	memcpy(hdr, mbuf, tpos);
	hdr[tpos] = '\0';

	char*	post = NULL;
	if (len > tpos)
	{
		post = new char[len - tpos];
		memcpy(post, mbuf + tpos, len - tpos);
	}
	delete[] mbuf;
	mbuf = post;
	len -= tpos;

	// decode header
	HTTP_RQ*	rq = new HTTP_RQ;
	rq->AttachHeaders(hdr);

	// get content length
	const char*	cl = rq->GetHeader("Content-Length");
	if (cl)
	{
		size_t datalen = atoi(cl);

		while (datalen > len)
		{
			int	rlen = Recv(s, buf, 512, 0);

			if (rlen == SOCKET_ERROR)
			{
				delete[] mbuf;
				delete rq;
				return NULL;
			}

			if (len)
			{
				// re-alloc
				char*	nbuf = new char[len + rlen + 1];
				memcpy(nbuf, mbuf, len);
				memcpy(nbuf + len, buf, rlen);
				delete[] mbuf;
				mbuf = nbuf;
				len += rlen;
			}
			else
			{
				mbuf = new char[rlen + 1];
				memcpy(mbuf, buf, rlen);
				len = rlen;
			}
		}

		// IE sometimes sends a spare <0D><0A> just to
		// make life interesting for server authors - if
		// you close the socket without receiving these
		// bytes it thinks you don't love it

		// wait 1ms for the data
		fd_set	set;

		FD_ZERO(&set);
		FD_SET(s, &set);

		struct timeval	tv;

		tv.tv_sec = 0;
		tv.tv_usec = 1000;

		if (select(1, &set, 0, 0, &tv))
		{
			char	buf[2];
			recv(s, buf, 2, 0);
		}
	}

	// attach data buffer
	rq->AttachData(mbuf, len);

	return rq;
}

void Server::SetCookie(Headers& hdrs, const char* key, const char* value) const
{
	char*	spec = new char[strlen(key) + strlen(value) + 32];

	sprintf(spec, "%s=%s; path=/;", key, value);

	hdrs.AddHeader("Set-Cookie", spec);

	delete[] spec;
}

void Server::RunCmd(SOCKET s, HTTP_RQ* rq)
{
	// create temp file
	TempFile	header("HDR");
	TempFile	body("BDY");

	// run command
	Headers		hdrs;
	int			status = 501;	// not implemented
	bool		senddata = true;

	if (0)
	{
		// log the LogID cookie; create if needed
		// note that I hate this idea but the code is
		// left in because it's a good example of using
		// the GetCookie and SetCookie functions
		const char*	cookie = rq->GetCookie("LogID");

		if (cookie)
		{
			// Log ID returned by browser
			Log(",LogID=%s", cookie);
		}
		else
		{
			// create a LogID and set it on the client
			char	cookie[16];
			sprintf(cookie, "%4.4X%4.4X", rand(), rand());

			SetCookie(hdrs, "LogID", cookie);

			Log(",LogID=%s", cookie);
		}
	}

	AddServerHeaders(hdrs);

	// log the request
	Log(",%s,%s", rq->GetReq(), rq->GetURI());

	// send the request to the subclass
	if (!stricmp(rq->GetReq(), "get"))
	{
		// GET operation
		status = Get(body, rq, hdrs);
	}
	if (!stricmp(rq->GetReq(), "head"))
	{
		// HEAD operation - return headers only
		status = Get(body, rq, hdrs);
		senddata = false;
	}
	else if (!stricmp(rq->GetReq(), "post"))
	{
		// POST operation
		//
		// IE5 does not distinguish POST from GET, so we do not either
		// POST/GET must be determined from the supplied parameters
		status = Get(body, rq, hdrs);
	}

	// close file
	body.Close();
	if (status != 200)
	{
		HTTPError(s, status);
		return;
	}

	// open data for reading
	FILE*	bfd = body.Open("rb");
	if (!bfd)
	{
		HTTPError(s, 500);
		return;
	}

	// get size
	fseek(bfd, 0, SEEK_END);
	size_t	blen = ftell(bfd);
	fseek(bfd, 0, SEEK_SET);

	size_t bsent = blen;

	// add Content-Length header
	char	temp[16];
	sprintf(temp, "%d", blen);
	hdrs.AddHeader("Content-Length", temp);

	// write headers
	FILE*	hfd = header.Open("w");
	if (!hfd)
	{
		HTTPError(s, 500);
		return;
	}

	// write the status and headers to file
	fprintf(hfd, "%s\n", StatusLine(status));
	hdrs.Write(hfd);
	fprintf(hfd, "\n");
	header.Close();

	// open headers for reading
	hfd = header.Open("rb");
	if (!hfd)
	{
		HTTPError(s, 500);
		return;
	}

	// get size
	fseek(hfd, 0, SEEK_END);
	size_t hlen = ftell(hfd);
	fseek(hfd, 0, SEEK_SET);

	char*	head = new char[hlen];
	fread(head, 1, hlen, hfd);
	header.Close();

	// send header
	Send(s, head, hlen, 0);

	delete[] head;

	// send file
	if (senddata && blen)
	{
		char*	data = new char[65536];

		while (blen)
		{
			// desired length
			size_t dlen = blen;
			if (dlen > 65536)	dlen = 65536;

			size_t rlen = fread(data, 1, dlen, bfd);
			if (rlen == 0)
			{
				// failed to read any
				delete[] data;
				Log(",%d,%d,%s", bsent, 500,StatusLine(500));
				return;
			}
			Send(s, data, rlen, 0);
			blen -= rlen;
		}

		delete[] data;
	}

	// log the status
	Log(",%d,%d,%s", bsent, status,StatusLine(status));
}
