// utility.cpp
//
// utilities

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "os.h"
#include "utility.h"

// CS_Compare
//
// compare two strings

int CS_Compare(const char* s1, const char* s2, bool ignore_case)
{
	return ignore_case ? stricmp(s1, s2) : strcmp(s1, s2);
}

// CS_islower
// CS_isupper
// CS_isalpha
// CS_isalnum
//
// for ISO-8859-1
// from http://www.utoronto.ca/webdocs/HTMLdocs/NewHTML/iso_table.html

bool CS_islower(char c)
{
	if ((c >= 'a') && (c <= 'z'))				return true;
	if (c == char(181))							return true;
	if (c == char(247))							return false;
	if ((c >= char(224)) && (c <= char(255)))	return true;
	return false;
}

bool CS_isupper(char c)
{
	if ((c >= 'A') && (c <= 'Z'))				return true;
	if (c == char(215))							return false;
	if ((c >= char(192)) && (c <= char(223)))	return true;
	return false;
}

bool CS_isalpha(char c)
{
	return CS_islower(c) || CS_isupper(c);
}

bool CS_isalnum(char c)
{
	if ((c >= '0') && (c <= '9'))	return true;
	return CS_isalpha(c);
}

bool CS_isdigit(char c)
{
	if ((c >= '0') && (c <= '9'))	return true;
	return false;
}

// StringSplit
//
// split text at character (at)
// frees text and returns new text buffer, plus new
// part buffer - user is responsible for freeing
// these buffers, and for not passing in static
// data!

char* StringSplit(char* text, char at, char*& part)
{
	char*	tptr = text;
	while (*tptr && (*tptr != at))	tptr++;
	if (!*tptr)
	{
		part = NULL;
		return text;
	}

	// check for [<0D>]<0A>
	if ((at == 0x0A) && (tptr > text) && (tptr[-1] == 0x0D))
	{
		part = new char[tptr - text];
		memcpy(part, text, tptr - text - 1);
		part[tptr - text - 1] = '\0';
	}
	else
	{
		part = new char[tptr - text + 1];
		memcpy(part, text, tptr - text);
		part[tptr - text] = '\0';
	}

	tptr++;

	char*	ntext = new char[strlen(tptr) + 1];
	strcpy(ntext, tptr);
	delete[] text;

	return ntext;
}

// StripSpaces
//
// Strip spaces from str

void StripSpaces(char* str)
{
	char* tmpstr = new char[strlen(str) + 1];
	unsigned int tmpstrpos = 0;
	
	for (unsigned int i=0; i<strlen(str)+1; i++)
	{
		if (str[i] != ' ')
		{
			tmpstr[tmpstrpos] = str[i];
			tmpstrpos++;
		}
	}

	strcpy(str, tmpstr);
}


// StringDup
//
// duplicate a string

char* StringDup(const char* str)
{
	char*	nstr = new char[strlen(str) + 1];
	strcpy(nstr, str);
	return nstr;
}

// StringDup
//
// duplicate a string

char* StringDup(const char* start, const char* end)
{
	char*	text = new char[end - start + 1];
	memcpy(text, start, end - start);
	text[end - start] = 0;
	return text;
}

// hex
//
// get the hex value of a character, else 0

static int hex(char c)
{
	if ((c >= 'A') && (c <= 'Z'))	return c - 'A' + 0x0A;
	if ((c >= 'a') && (c <= 'z'))	return c - 'a' + 0x0a;
	if ((c >= '0') && (c <= '9'))	return c - '0' + 0x00;

	return 0;
}

// URLEncode
// URLDecode
//
// encode/decode data from URL mapping
//
// functions delete input string and return new output buffer

void URLDecode(char*& data)
{
	char*	obuf = new char[strlen(data) + 1];
	char*	optr = obuf;

	char*	dptr = data;

	while (*dptr)
	{
		if (*dptr == '+')
		{
			*optr++ = ' ';
			dptr++;
		}
		else if (*dptr == '%')
		{
			if (!dptr[1] || !dptr[2])	break;
			*optr++ = hex(dptr[1]) * 0x10 + hex(dptr[2]);
			dptr += 3;
		}
		else
		{
			*optr++ = *dptr++;
		}
	}

	*optr++ = 0;
	delete[] data;
	data = obuf;
}

void URLEncode(char*& data)
{
	static char* hdig = "0123456789abcdef";

	char*	obuf = new char[strlen(data) * 3 + 1];
	char*	optr = obuf;

	char*	dptr = data;

	while (*dptr)
	{
		if (((*dptr >= 'a') && (*dptr <= 'z')) ||
			((*dptr >= 'A') && (*dptr <= 'Z')) ||
			((*dptr >= '0') && (*dptr <= '9')) ||
			(*dptr == '/') || (*dptr == '-') || (*dptr == ':'))
		{
			*optr++ = *dptr++;
		}
		else
		{
			*optr++ = '%';
			*optr++ = hdig[(*dptr >> 4) & 15];
			*optr++ = hdig[*dptr & 15];
			dptr++;
		}
	}
	*optr++ = '\0';
	delete[] data;
	data = obuf;
}

// UTFDecode
//
// decode from UTF-8 into ISO-8859-1
// this is a simple subset of UTF-8:
//
// C2 XX    -> (XX & 3F) + 80
// C3 XX    -> (XX & 3F) + C0
//
// conversion is in-place (it can only shrink the string)

void UTFDecode(char* data)
{
	const char*	iptr = data;
	char*		optr = data;

	while (*iptr)
	{
		if ((iptr[0] >= char(0xC0)) && (iptr[0] <= char(0xC3)))
		{
			char	n = char((iptr[1] & 0x3F) + ((iptr[0] << 6) & 0xC0));

			*optr++ = n;
			iptr += 2;
		}
		else
		{
			*optr++ = *iptr++;
		}
	}

	*optr++ = 0;
}

// UTFEncode
//
// encode from ISO-8859-1 to UTF-8
//
// function deletes input string and returns new output buffer

void UTFEncode(char*& data)
{
	char*		obuf = new char[strlen(data) * 2 + 1];
	char*		optr = obuf;
	const char*	iptr = data;

	while (*iptr)
	{
		if ((iptr[0] >= char(0x80)) && (iptr[0] <= char(0xBF)))
		{
			*optr++ = char(0xC2);
			*optr++ = (iptr[0] & 0x3F) | 0x80;
		}
		else if ((iptr[0] >= char(0xC0)) && (iptr[0] <= char(0xFF)))
		{
			*optr++ = char(0xC3);
			*optr++ = (iptr[0] & 0x3F) | 0x80;
		}
		else
		{
			*optr++ = iptr[0];
		}
		iptr++;
	}
	*optr++ = 0;
	delete[] data;
	data = obuf;
}

// GetX
//
// decode an item from a string; return newly-allocated buffer
// and advance ls past the item's end

// get line terminated by [<CR>]<LF>

char* GetLine(const char*& ls)
{
	const char*	le = ls;
	while (*le && (*le != '\n'))		le++;

	char* line = StringDup(ls, le);

	// remove <CR>
	if ((le > ls) && (le[-1] == 0x0D))	line[le - ls - 1] = 0;
	if (*le)	le++;
	ls = le;

	return line;
}

// get an alphanumeric string

char* GetAlphaNumeric(const char*& ls)
{
	const char*	le = ls;
	while (CS_isalnum(*le))	le++;

	char* word = StringDup(ls, le);

	ls = le;

	return word;
}

// get valid URL

static bool isurl(char c)
{
	// leaf characters
	if (isalnum(c) || (c == '-') || (c == '_'))	return true;
	// URL seperators
	if ((c == ':') || (c == '/') || (c == '\\') || (c == '.') || (c == '#'))	return true;
	// CGI access characters
	if ((c == '+') || (c == '%') || (c == '?') || (c == '&') || (c == '='))	return true;
	// email characters
	if (c == '@')	return true;
	// UNIX special characters
	if (c == '~')	return true;

	return false;
}

char* GetURL(const char*& ls)
{
	const char*	le = ls;
	while (isurl(*le))	le++;

	if (le > ls)
	{
		// check last character isn't punctuation
		if ((le[-1] == '.') || (le[-1] == '?'))	le--;
	}

	// copy URL
	char*	url = StringDup(ls, le);

	ls = le;
	
	return url;
}

// get a decimal number

char* GetNumber(const char*& ls)
{
	const char*	le = ls;
	while (CS_isdigit(*le))	le++;

	char*	num = StringDup(ls, le);

	ls = le;

	return num;
}

// get ISBN - ignore leading spaces, allow - anywhere,
// last digit can be X ... returns NULL if invalid
//
// also copies raw ISBN (just digits) into rawisbn

char* GetISBN(const char*& ls, char rawisbn[11])
{
	const char* le = ls;
	if (*le != ' ')	return NULL;
	le++;

	int		ix = 0;

	for (; *le; le++)
	{
		if (*le == '-')		continue;

		if ((*le >= '0') && (*le <= '9'))
		{
			rawisbn[ix] = *le;
			ix++;
			if (ix == 11)	return NULL;
		}
		else if ((*le == 'X') || (*le == 'x'))
		{
			if (ix != 9)	return NULL;
			rawisbn[ix] = *le;
			ix++;
		}
		else break;
	}

	if (ix != 10)	return NULL;
	rawisbn[10] = 0;

	char*	isbn = StringDup(ls, le);

	ls = le;

	return isbn;
}

// MonthName
//
// 1 = January, ..., 12 = December

const char* MonthName(int ix)
{
	static char* month[] = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };

	if ((ix < 1) || (ix > 12))	return "";
	
	return month[ix - 1];
}

// MatchWildcardString
//
// see if a wildcard string is within another string

bool MatchWildcardString(const char* string, const char* search, bool recursing)
{
	// scan along string
	while (string[0])
	{
		if (recursing && (string[0] == '\n'))	return false;	// don't let * go over \n

		const char*	sptr = string;
		const char*	wptr = search;

		// scan along search string
		while (*wptr)
		{
			if (*wptr == '*')	return MatchWildcardString(sptr, wptr + 1, true);
			if (((*wptr == '?') && (*sptr != '\n')) ||			// ? matches anything but \n
				(toupper(*wptr) == toupper(*sptr)))				// not case sensitive
			{
				wptr++;
				sptr++;
			}
			else break;
		}

		if (!*wptr)	return true;

		string++;
	}

	return false;
}

// DecodeBase64
//
// decode a Base64 string [see RFC 1341]

static char base64_table[256] =
{
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,
	52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-2,-1,-1,
	-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
	15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,
	-1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,
	41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
};

char* DecodeBase64(const char* input, int& len)
{
	char*	buffer = new char[strlen(input) + 1];
	int		pos = 0;
	
	while (*input)
	{
		int	a = base64_table[int(input[0]) & 255];
		if (a == -1)	break;
		int	b = base64_table[int(input[1]) & 255];
		if (b == -1)	break;
		int	c = base64_table[int(input[2]) & 255];
		if (c == -1)	break;
		int	d = base64_table[int(input[3]) & 255];
		if (d == -1)	break;

		input += 4;

		if (d != -2)
		{
			int	out = (a << 18) | (b << 12) | (c << 6) | d;

			buffer[pos++] = out >> 16;
			buffer[pos++] = out >> 8;
			buffer[pos++] = out;
		}
		else if (c != -2)
		{
			int	out = (a << 18) | (b << 12) | (c << 6);

			buffer[pos++] = out >> 16;
			buffer[pos++] = out >> 8;
		}
		else
		{
			int	out = (a << 18) | (b << 12);

			buffer[pos++] = out >> 16;
		}
	}
	
	buffer[pos] = 0;
	len = pos;

	return buffer;
}

// NOT FULLY TESTED

char* EncodeBase64(const char* input, int& len)
{
	static char	base64_encode[66] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcedfghijklmnopqrstuvwxyz0123456789+/=";

	len = (strlen(input) + 2) / 3;		// number of 3-char sequences in input
	len *= 4;							// output len

	char*	buffer = new char[len + 1];
	int		pos = 0;

	while (*input)
	{
		int	a,b,c,d;

		if (input[1])
		{
			if (input[2])
			{
				// 3 chars
				int	out = (int(input[0]) << 16) | (int(input[1] << 8)) | int(input[2]);

				d = out & 63;	out >>= 6;
				c = out & 63;	out >>= 6;
				b = out & 63;	out >>= 6;
				a = out & 63;

				input += 3;
			}
			else
			{
				// 2 chars
				int	out = (int(input[0]) << 16) | (int(input[1] << 8));

				d = 64;			out >>= 6;
				c = out & 63;	out >>= 6;
				b = out & 63;	out >>= 6;
				a = out & 63;

				input += 2;
			}
		}
		else
		{
			// 1 char
			int out = int(input[0] << 16);

			d = 64;			out >>= 6;
			c = 64;			out >>= 6;
			b = out & 63;	out >>= 6;
			a = out & 63;
		}

		buffer[pos++] = base64_encode[a];
		buffer[pos++] = base64_encode[b];
		buffer[pos++] = base64_encode[c];
		buffer[pos++] = base64_encode[d];
	}

	buffer[pos] = 0;

	return buffer;
}

// GetSysTime
//
// get the current time in system format as a (static) string

const char* GetSysTime()
{
	static char systime[64];

	time_t		tnow;

	time(&tnow);
	strftime(systime, 64, "%H:%M:%S", localtime(&tnow));
	
	return systime;
}

// GetSysDate
//
// get the current date in system format as a (static) string

const char* GetSysDate()
{
	static char sysdate[64];

	time_t		tnow;

	time(&tnow);
	strftime(sysdate, 64, "%a %b %d %Y", localtime(&tnow));

	return sysdate;
}

// LoadFile
//
// load a file into a new buffer

char* LoadFile(const char* filename)
{
	FILE*	fd = fopen(filename, "rb");
	if (!fd)	return 0;

	fseek(fd, 0, SEEK_END);
	size_t len = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	char* data = new char[len + 1];
	if (fread(data, 1, len, fd) != len)
	{
		delete[] data;
		fclose(fd);
		return 0;
	}
	fclose(fd);
	data[len] = 0;

	return data;
}


// Capitalise
//
// Makes each word start with a capital

void Capitalise(char* str)
{
	int i;
	bool space = true;

	for (i=0; i<(int)strlen(str);i++)
	{
		if (space && CS_islower( str[i] ))
		{
			str[i] -= 32;
		}

		if (str[i] == ' ')
		{
			space = true;
		}
		else
		{
			space = false;
		}
	}
}

// IsValidTitle
//
// is the string a valid title?

bool IsValidTitle(const char* tptr)
{
	// Must be <cap><small>
	if (!CS_isupper(*tptr))	return false;
	tptr++;
	if (!CS_islower(*tptr))	return false;
	tptr++;

	// Skip past all <small>
	while (*tptr)
	{
		if (CS_islower(*tptr))	tptr++;
		else					break;
	}

	// Now <cap> or <numeral>
	if (!CS_isupper(*tptr) && !CS_isdigit(*tptr))
	{
		return false;
	}

	// And no invalid rubbish afterwards
	while (*++tptr)
	{
		if (!CS_isalnum(*tptr))	return false;
	}

	// It's OK
	return true;
}

// IsAcceptableTitle
//
// is the string an acceptable title?

bool IsAcceptableTitle(const char* tptr)
{
	for (;*tptr;tptr++)
	{
		if (!CS_isalnum(*tptr))	return false;
	}

	return true;
}

// MakeSpacedTitle
//
// make a wiki name into a spaced wiki name

char* MakeSpacedTitle(const char* title)
{
	// if next letter is CAP and last letter was small or digit, insert space
	// if next letter is CAP and letter after it is small, insert space
	// if next letter is digit and last letter was small, insert space
	const char*	tptr = title;
	char* obuf = new char[strlen(title) * 2 + 1];
	char* optr = obuf;

	*optr++ = *tptr++;

	while (*tptr)
	{
		if (CS_isupper(*tptr))
		{
			// aA -> a A
			// 0A -> 0 A
			// AA -> AA
			// AAa -> A Aa
			if (CS_islower(tptr[-1]))		*optr++ = ' ';
			else if (CS_isdigit(tptr[-1]))	*optr++ = ' ';
			else if (CS_islower(tptr[1]))	*optr++ = ' ';
		}
		else if (CS_isdigit(*tptr))
		{
			// a0 -> a 0
			// A0 -> A0
			// 00 -> 00
			if (CS_islower(tptr[-1]))			*optr++ = ' ';
		}
		*optr++ = *tptr++;
	}
	*optr++ = *tptr++;

	return obuf;
}
